from typing import TYPE_CHECKING, Any

import rust_ephem

from ..common import (
    ACSCommandType,
    ACSMode,
    dtutcfromtimestamp,
    unixtime2date,
    unixtime2yearday,
)
from ..config import MissionConfig
from ..config.constants import DTOR
from ..simulation.passes import PassTimes
from .acs_command import ACSCommand
from .emergency_charging import EmergencyCharging
from .passes import Pass
from .slew import Slew

if TYPE_CHECKING:
    from ..ditl.ditl_log import DITLLog
    from ..targets import Pointing


class ACS:
    """
    Queue-driven state machine for spacecraft Attitude Control System (ACS).

    The ACS manages spacecraft pointing through a command queue, where each command
    represents a state transition (slew, pass, return to pointing, etc.). The state
    machine processes commands at their scheduled execution times and maintains
    current pointing state.
    """

    ephem: rust_ephem.Ephemeris
    slew_dists: list[float]
    ra: float
    dec: float
    roll: float
    obstype: str
    acsmode: ACSMode
    command_queue: list[ACSCommand]
    executed_commands: list[ACSCommand]
    current_slew: Slew | None
    last_ppt: Slew | None
    last_slew: Slew | None
    in_eclipse: bool

    def __init__(self, config: MissionConfig, log: "DITLLog | None" = None) -> None:
        """Initialize the Attitude Control System.

        Args:
            constraint: Constraint object with ephemeris.
            config: MissionConfiguration object.
            log: Optional DITLLog for event logging. If None, prints to stdout.
        """
        assert config.constraint is not None, "Constraint must be provided to ACS"
        self.constraint = config.constraint
        self.config = config
        self.log = log

        # Configuration
        assert self.constraint.ephem is not None, "Ephemeris must be set in Constraint"
        self.ephem = self.constraint.ephem

        # Initial pointing derived from ephemeris (opposite Earth vector)
        self.ra = (180 + self.ephem.earth[0].ra.deg) % 360
        self.dec = -self.ephem.earth[0].dec.deg
        # Set up initial last_slew pointing (this would have been the
        # last slew to execute before the current DITL), so didn't
        # happen in our simulation, but defines a realistic boundary
        # condition for our simulation.
        self.last_slew = Slew(
            config=config,
        )
        self.last_slew.endra = self.ra
        self.last_slew.enddec = self.dec

        # Current state
        self.roll = 0.0
        self.obstype = "PPT"
        self.acsmode = ACSMode.SCIENCE  # Start in science/pointing mode
        self.in_eclipse = False  # Initialize eclipse state
        self.in_safe_mode = False  # Safe mode flag - once True, cannot be exited

        # Command queue (sorted by execution_time)
        self.command_queue = []
        self.executed_commands = []

        # Current and historical state
        self.current_slew = None
        self.last_ppt = None

        self.passrequests = PassTimes(config=config)
        self.current_pass: Pass | None = None
        self.solar_panel = config.solar_panel
        self.slew_dists: list[float] = []
        self.saa = None

    def _log_or_print(self, utime: float, event_type: str, description: str) -> None:
        """Log an event to DITLLog if available, otherwise print to stdout.

        Args:
            utime: Unix timestamp.
            event_type: Event category (ACS, SLEW, PASS, etc.).
            description: Human-readable description.
        """
        if self.log is not None:
            self.log.log_event(
                utime=utime,
                event_type=event_type,
                description=description,
                obsid=getattr(self.last_slew, "obsid", None)
                if self.last_slew
                else None,
                acs_mode=self.acsmode if hasattr(self, "acsmode") else None,
            )
        else:
            # Fallback to print if no log available
            print(description)

    def enqueue_command(self, command: ACSCommand) -> None:
        """Add a command to the queue, maintaining time-sorted order.

        Commands cannot be enqueued if safe mode has been entered, except
        for SAFE slews which are part of safe mode entry.
        """
        # Allow SAFE slews to be enqueued even in safe mode (part of safe mode entry)
        is_safe_slew = (
            command.command_type == ACSCommandType.SLEW_TO_TARGET
            and hasattr(command, "slew")
            and command.slew is not None
            and command.slew.obstype == "SAFE"
        )

        # Prevent any commands from being enqueued in safe mode (except SAFE slews)
        if self.in_safe_mode and not is_safe_slew:
            self._log_or_print(
                command.execution_time,
                "ACS",
                f"{unixtime2date(command.execution_time)}: Command {command.command_type.name} rejected - spacecraft is in SAFE MODE",
            )
            return

        self.command_queue.append(command)
        self.command_queue.sort(key=lambda cmd: cmd.execution_time)
        self._log_or_print(
            command.execution_time,
            "ACS",
            f"{unixtime2date(command.execution_time)}: Enqueued {command.command_type.name} command for execution  (queue size: {len(self.command_queue)})",
        )

    def _process_commands(self, utime: float) -> None:
        """Process all commands scheduled for execution at or before current time."""
        while self.command_queue and self.command_queue[0].execution_time <= utime:
            command = self.command_queue.pop(0)
            self._log_or_print(
                utime,
                "ACS",
                f"{unixtime2date(utime)}: Executing {command.command_type.name} command.",
            )

            # Dispatch to appropriate handler based on command type
            handlers: dict[ACSCommandType, Any] = {
                ACSCommandType.SLEW_TO_TARGET: lambda: self._handle_slew_command(
                    command, utime
                ),
                ACSCommandType.START_PASS: lambda: self._start_pass(command, utime),
                ACSCommandType.END_PASS: lambda: self._end_pass(utime),
                ACSCommandType.START_BATTERY_CHARGE: lambda: self._start_battery_charge(
                    command, utime
                ),
                ACSCommandType.END_BATTERY_CHARGE: lambda: self._end_battery_charge(
                    utime
                ),
                ACSCommandType.ENTER_SAFE_MODE: lambda: self._handle_safe_mode_command(
                    utime
                ),
            }

            handler = handlers.get(command.command_type)
            if handler:
                handler()
            self.executed_commands.append(command)

    def _handle_slew_command(self, command: ACSCommand, utime: float) -> None:
        """Handle SLEW_TO_TARGET command."""
        if command.slew is not None:
            self._start_slew(command.slew, utime)

    # Handle Ground Station Pass Commands
    def _start_pass(self, command: ACSCommand, utime: float) -> None:
        """Handle START_PASS command to command the start of a groundstation pass."""
        # Fetch the current pass from pass requests
        self.current_pass = self.passrequests.current_pass(utime)
        if self.current_pass is None:
            self._log_or_print(
                utime, "PASS", f"{unixtime2date(utime)}: No active pass found to start."
            )
            return
        self.acsmode = ACSMode.PASS
        self._log_or_print(
            utime,
            "PASS",
            f"{unixtime2date(utime)}: Starting pass over groundstation {self.current_pass.station}.",
        )

    def _end_pass(self, utime: float) -> None:
        """Handle the END_PASS command to command the end of a groundstation pass."""
        self.current_pass = None
        self.acsmode = ACSMode.SCIENCE

        self._log_or_print(
            utime,
            "PASS",
            f"{unixtime2date(utime)}: Pass over - returning to last PPT {getattr(self.last_ppt, 'obsid', 'unknown')}",
        )

    # Handle Safe Mode Command
    def _handle_safe_mode_command(self, utime: float) -> None:
        """Handle ENTER_SAFE_MODE command.

        Once safe mode is entered, it cannot be exited. The spacecraft will
        point solar panels at the Sun and obey bus-level constraints.
        """
        self._log_or_print(
            utime, "SAFE", f"{unixtime2date(utime)}: Entering SAFE MODE - irreversible"
        )
        self.in_safe_mode = True
        # Clear command queue to prevent any future commands from executing
        self.command_queue.clear()
        self._log_or_print(
            utime,
            "SAFE",
            f"{unixtime2date(utime)}: Command queue cleared - no further commands will be executed",
        )

        # Initiate slew to Sun pointing for safe mode
        # Use solar panel optimal pointing to maximize power generation
        if self.solar_panel is not None:
            safe_ra, safe_dec = self.solar_panel.optimal_charging_pointing(
                utime, self.ephem
            )
        else:
            # Fallback: point directly at Sun if no solar panel config
            index = self.ephem.index(dtutcfromtimestamp(utime))
            safe_ra = self.ephem.sun[index].ra.deg
            safe_dec = self.ephem.sun[index].dec.deg

        self._log_or_print(
            utime,
            "SAFE",
            f"{unixtime2date(utime)}: Initiating slew to safe mode pointing at RA={safe_ra:.2f} Dec={safe_dec:.2f}",
        )
        # Enqueue slew to safe pointing with a special obsid
        self._enqueue_slew(safe_ra, safe_dec, obsid=-999, utime=utime, obstype="SAFE")

    def _start_slew(self, slew: Slew, utime: float) -> None:
        """Start executing a slew.

        The ACS always drives the spacecraft from its current position. When a slew
        command is executed, we set the start position to the current ACS pointing
        and recalculate the slew profile. This ensures continuous motion without
        teleportation, regardless of when commands were originally scheduled.
        """
        # Always start slew from current spacecraft position - ACS drives the spacecraft
        slew.startra = self.ra
        slew.startdec = self.dec
        slew.slewstart = utime
        slew.calc_slewtime()

        self._log_or_print(
            utime,
            "SLEW",
            f"{unixtime2date(utime)}: Starting slew from RA={self.ra:.2f} Dec={self.dec:.2f} "
            f"to RA={slew.endra:.2f} Dec={slew.enddec:.2f} (duration: {slew.slewtime:.1f}s)",
        )

        self.current_slew = slew
        self.last_slew = slew

        # Update last_ppt if this is a science pointing
        if self._is_science_pointing(slew):
            self.last_ppt = slew

    def _is_science_pointing(self, slew: Slew) -> bool:
        """Check if slew represents a science pointing (not a pass)."""
        return slew.obstype == "PPT" and isinstance(slew, Slew)

    def _enqueue_slew(
        self, ra: float, dec: float, obsid: int, utime: float, obstype: str = "PPT"
    ) -> bool:
        """Create and enqueue a slew command.

        This is a private helper method used internally by ACS for creating slew
        commands during battery charging operations.
        """
        # Create slew object
        slew = Slew(
            config=self.config,
        )
        slew.ephem = self.ephem
        slew.slewrequest = utime
        slew.endra = ra
        slew.enddec = dec
        slew.obstype = obstype
        slew.obsid = obsid

        # For SAFE mode, skip visibility checking (emergency situation)
        if obstype == "SAFE":
            # Initialize slew positions without target
            is_first_slew = self._initialize_slew_positions(slew, utime)
            slew.at = None  # No visibility constraint in safe mode
            execution_time = utime  # Execute immediately
        else:
            # Set up target observation request and check visibility
            target_request = self._create_target_request(slew, utime)
            slew.at = target_request

            visstart = target_request.next_vis(utime)
            is_first_slew = self._initialize_slew_positions(slew, utime)

            # Validate slew is possible
            if not self._is_slew_valid(visstart, slew.obstype, utime):
                return False

            # Calculate slew timing
            execution_time = self._calculate_slew_timing(
                slew, visstart, utime, is_first_slew
            )

        slew.slewstart = execution_time
        slew.calc_slewtime()
        self.slew_dists.append(slew.slewdist)

        # Enqueue the slew command
        command = ACSCommand(
            command_type=ACSCommandType.SLEW_TO_TARGET,
            execution_time=slew.slewstart,
            slew=slew,
        )
        self.enqueue_command(command)

        if is_first_slew:
            self.last_slew = slew

        return True

    def _create_target_request(self, slew: Slew, utime: float) -> "Pointing":
        """Create and configure a target observation request for visibility checking."""
        from ..targets import Pointing

        target = Pointing(
            config=self.config,
            ra=slew.endra,
            dec=slew.enddec,
            obsid=slew.obsid,
        )
        target.isat = slew.obstype != "PPT"

        year, day = unixtime2yearday(utime)
        target.visibility()
        return target

    def _initialize_slew_positions(self, slew: Slew, utime: float) -> bool:
        """Initialize slew start positions.

        If a previous slew exists, start from current pointing (self.ra/dec).
        If this is the first slew, derive current pointing from ephemeris if
        ra/dec have not yet been initialized (both zero) and use that as start.
        Returns True if this is the first slew (used for accounting/logging).
        """
        if self.last_slew:
            slew.startra = self.ra
            slew.startdec = self.dec
            return False

        slew.startra = self.ra
        slew.startdec = self.dec
        return True

    def _is_slew_valid(self, visstart: float, obstype: str, utime: float) -> bool:
        """Check if the requested slew is valid (target is visible)."""
        if not visstart and obstype == "PPT":
            self._log_or_print(
                utime,
                "SLEW",
                f"{unixtime2date(utime)}: Slew rejected - target not visible",
            )
            return False
        return True

    def _calculate_slew_timing(
        self, slew: Slew, visstart: float, utime: float, is_first_slew: bool
    ) -> float:
        """Calculate when the slew should start, accounting for current slew and constraints."""
        execution_time = utime

        # Wait for current slew to finish if in progress
        if (
            not is_first_slew
            and isinstance(self.last_slew, Slew)
            and self.last_slew.is_slewing(utime)
        ):
            execution_time = self.last_slew.slewstart + self.last_slew.slewtime
            self._log_or_print(
                utime,
                "SLEW",
                "%s: Slewing - delaying next slew until %s"
                % (
                    unixtime2date(utime),
                    unixtime2date(execution_time),
                ),
            )

        # Wait for target visibility if constrained
        if visstart > execution_time and slew.obstype == "PPT":
            self._log_or_print(
                utime,
                "SLEW",
                "%s: Slew delayed by %.1fs"
                % (
                    unixtime2date(utime),
                    visstart - execution_time,
                ),
            )
            execution_time = visstart

        return execution_time

    def pointing(self, utime: float) -> tuple[float, float, float, int]:
        """
        Calculate ACS pointing for the given time.

        This is the main state machine update method. It:
        1. Checks for upcoming passes and enqueues commands
        2. Processes any commands due for execution
        3. Updates the current ACS mode based on slew/pass state
        4. Calculates current RA/Dec pointing
        """
        # Determine if the spacecraft is currently in eclipse
        self.in_eclipse = self.constraint.in_eclipse(ra=0, dec=0, time=utime)

        # Process any commands scheduled for execution at or before current time
        self._process_commands(utime)

        # Update ACS mode based on current state
        self._update_mode(utime)

        # Check current constraints
        self._check_constraints(utime)

        # Calculate current RA/Dec pointing
        self._calculate_pointing(utime)

        # Calculate roll angle
        # FIXME: Rolls should be pre-calculated, as this is computationally expensive
        if False:
            from ..simulation.roll import optimum_roll

            self.roll = optimum_roll(
                self.ra * DTOR,
                self.dec * DTOR,
                utime,
                self.ephem,
                self.solar_panel,
            )

        # Return current pointing
        if self.last_slew is not None:
            return self.ra, self.dec, self.roll, self.last_slew.obsid
        else:
            return self.ra, self.dec, self.roll, 1

    def get_mode(self, utime: float) -> ACSMode:
        """Determine current spacecraft mode based on ACS state and external factors.

        This is the authoritative source for determining spacecraft operational mode,
        considering slewing state, passes, SAA region, battery charging, and safe mode.
        """
        # Safe mode takes absolute priority - once entered, cannot be exited
        if self.in_safe_mode:
            return ACSMode.SAFE

        # Check if actively slewing
        if self._is_actively_slewing(utime):
            assert self.current_slew is not None, (
                "Current slew must be set when actively slewing"
            )
            # Check if slewing for charging - but only report CHARGING if in sunlight
            if self.current_slew.obstype == "CHARGE":
                # Check eclipse state - no point being in CHARGING mode during eclipse
                if self.in_eclipse:
                    return ACSMode.SLEWING  # In eclipse, treat as normal slew
                return ACSMode.CHARGING
            return (
                ACSMode.PASS if self.current_slew.obstype == "GSP" else ACSMode.SLEWING
            )

        # Check if dwelling in charging mode (after slew to charge pointing)
        if self._is_in_charging_mode(utime):
            return ACSMode.CHARGING

        # Check if in pass dwell phase (after slew, during communication)
        if self._is_in_pass_dwell(utime):
            return ACSMode.PASS

        # Check if in SAA region
        if self.saa is not None and self.saa.insaa(utime):
            return ACSMode.SAA

        return ACSMode.SCIENCE

    def _is_actively_slewing(self, utime: float) -> bool:
        """Check if spacecraft is currently executing a slew."""
        return self.current_slew is not None and self.current_slew.is_slewing(utime)

    def _is_in_charging_mode(self, utime: float) -> bool:
        """Check if spacecraft is in charging mode (dwelling at charge pointing).

        Charging mode persists after slew completes until END_BATTERY_CHARGE command.
        Returns False during eclipse since charging is not useful without sunlight.
        """
        # Must have completed a CHARGE slew and not be actively slewing
        if not (
            self.last_slew is not None
            and self.last_slew.obstype == "CHARGE"
            and not self._is_actively_slewing(utime)
        ):
            return False

        # Check if spacecraft is in sunlight (not in eclipse)
        if self.ephem is None:
            # No ephemeris, assume sunlight (charging possible)
            return True

        # Only charging mode if NOT in eclipse
        return not self.in_eclipse

    def _is_in_pass_dwell(self, utime: float) -> bool:
        """Check if spacecraft is in pass dwell phase (stationary during groundstation contact)."""
        if self.current_pass is None:
            return False
        if self.current_pass.in_pass(utime):
            return True
        return False

    def _update_mode(self, utime: float) -> None:
        """Update ACS mode based on current slew/pass state."""
        self.acsmode = self.get_mode(utime)

    def _check_constraints(self, utime: float) -> None:
        """Check and log constraint violations for current pointing."""
        if (
            isinstance(self.last_slew, Slew)
            and self.last_slew.at is not None
            and not isinstance(self.last_slew.at, bool)
            and self.last_slew.obstype == "PPT"
            and self.constraint.in_constraint(
                self.last_slew.at.ra, self.last_slew.at.dec, utime
            )
        ):
            assert self.last_slew.at is not None

            # Collect only the true constraints
            true_constraints = []
            if self.last_slew.at.in_moon(utime):
                true_constraints.append("Moon")
            if self.last_slew.at.in_sun(utime):
                true_constraints.append("Sun")
            if self.last_slew.at.in_earth(utime):
                true_constraints.append("Earth")
            if self.last_slew.at.in_panel(utime):
                true_constraints.append("Panel")

            # Print only if there are true constraints
            if true_constraints:
                self._log_or_print(
                    utime,
                    "CONSTRAINT",
                    "%s: CONSTRAINT: RA=%s Dec=%s obsid=%s %s"
                    % (
                        unixtime2date(utime),
                        self.last_slew.at.ra,
                        self.last_slew.at.dec,
                        self.last_slew.obsid,
                        " ".join(true_constraints),
                    ),
                )
            # Note: acsmode remains SCIENCE - the DITL will decide if charging is needed

    def _calculate_pointing(self, utime: float) -> None:
        """Calculate current RA/Dec based on slew state or safe mode."""
        # Safe mode overrides all other pointing
        if self.in_safe_mode:
            self._calculate_safe_mode_pointing(utime)
        # If we are in a groundstations pass
        elif self.current_pass is not None:
            self.ra, self.dec = self.current_pass.ra_dec(utime)  # type: ignore[assignment]
        # If we are actively slewing
        elif self.last_slew is not None:
            self.ra, self.dec = self.last_slew.ra_dec(utime)
        else:
            # If there's no slew or pass, maintain current pointing
            pass

    def _calculate_safe_mode_pointing(self, utime: float) -> None:
        """Calculate safe mode pointing - point solar panels at the Sun.

        In safe mode, the spacecraft points to maximize solar panel illumination.
        This may be perpendicular to the Sun for side-mounted panels or directly
        at the Sun for body-mounted panels, following the optimal charging pointing.
        """
        # Use solar panel optimal pointing if available
        if self.solar_panel is not None:
            target_ra, target_dec = self.solar_panel.optimal_charging_pointing(
                utime, self.ephem
            )
        else:
            # Fallback: point directly at Sun if no solar panel config and that
            # serves you right for not having solar panels!
            index = self.ephem.index(dtutcfromtimestamp(utime))
            target_ra = self.ephem.sun[index].ra.deg
            target_dec = self.ephem.sun[index].dec.deg

        # If actively slewing to safe mode position, use slew interpolation
        if (
            self.current_slew is not None
            and self.current_slew.obstype == "SAFE"
            and self.current_slew.is_slewing(utime)
        ):
            self.ra, self.dec = self.current_slew.ra_dec(utime)
        else:
            # After slew completes or for continuous tracking, maintain optimal pointing
            self.ra = target_ra
            self.dec = target_dec

    def request_pass(self, gspass: Pass) -> None:
        """Request a groundstation pass."""
        # Check for overlap with existing passes
        for existing_pass in self.passrequests.passes:
            if self._passes_overlap(gspass, existing_pass):
                self._log_or_print(
                    gspass.begin, "ERROR", "ERROR: Pass overlap detected: %s" % gspass
                )
                return

        self.passrequests.passes.append(gspass)
        self._log_or_print(gspass.begin, "PASS", "Pass requested: %s" % gspass)

    def _passes_overlap(self, pass1: Pass, pass2: Pass) -> bool:
        """Check if two passes have overlapping time windows."""
        # Passes overlap if one starts before the other ends
        return not (pass1.end <= pass2.begin or pass1.begin >= pass2.end)

    def request_battery_charge(
        self, utime: float, ra: float, dec: float, obsid: int
    ) -> None:
        """Request emergency battery charging at specified pointing.

        Enqueues a START_BATTERY_CHARGE command with the given pointing parameters.
        The command will be executed at the specified time.
        """
        command = ACSCommand(
            command_type=ACSCommandType.START_BATTERY_CHARGE,
            execution_time=utime,
            ra=ra,
            dec=dec,
            obsid=obsid,
        )
        self.enqueue_command(command)
        self._log_or_print(
            utime,
            "CHARGING",
            f"Battery charge requested at RA={ra:.2f} Dec={dec:.2f} obsid={obsid}",
        )

    def request_end_battery_charge(self, utime: float) -> None:
        """Request termination of emergency battery charging.

        Enqueues an END_BATTERY_CHARGE command to be executed at the specified time.
        """
        command = ACSCommand(
            command_type=ACSCommandType.END_BATTERY_CHARGE,
            execution_time=utime,
        )
        self.enqueue_command(command)
        self._log_or_print(utime, "CHARGING", "End battery charge requested")

    def request_safe_mode(self, utime: float) -> None:
        """Request entry into safe mode.

        Enqueues an ENTER_SAFE_MODE command to be executed at the specified time.
        Once safe mode is entered, it cannot be exited. The spacecraft will point
        its solar panels at the Sun and obey bus-level constraints.

        Warning: This is an irreversible operation.
        """
        command = ACSCommand(
            command_type=ACSCommandType.ENTER_SAFE_MODE,
            execution_time=utime,
        )
        self.enqueue_command(command)
        self._log_or_print(
            utime,
            "SAFE",
            f"{unixtime2date(utime)}: Safe mode entry requested - this is irreversible",
        )

    def initiate_emergency_charging(
        self,
        utime: float,
        ephem: rust_ephem.Ephemeris,
        emergency_charging: EmergencyCharging,
        lastra: float,
        lastdec: float,
        current_ppt: "Pointing | None",
    ) -> tuple[float, float, Any]:
        """Initiate emergency charging by creating charging PPT and enqueuing charge command.

        Delegates to EmergencyCharging module to create the optimal charging pointing,
        then automatically enqueues the battery charge command via request_battery_charge().

        Returns:
            Tuple of (new_ra, new_dec, charging_ppt) where charging_ppt is the
            created charging pointing or None if charging could not be initiated.
        """
        charging_ppt = emergency_charging.initiate_emergency_charging(
            utime, ephem, lastra, lastdec, current_ppt
        )
        if charging_ppt is not None:
            self.request_battery_charge(
                utime, charging_ppt.ra, charging_ppt.dec, charging_ppt.obsid
            )
            return charging_ppt.ra, charging_ppt.dec, charging_ppt
        return lastra, lastdec, None

    def _start_battery_charge(self, command: ACSCommand, utime: float) -> None:
        """Handle START_BATTERY_CHARGE command execution.

        Initiates a slew to the optimal charging pointing.
        """
        if (
            command.ra is not None
            and command.dec is not None
            and command.obsid is not None
        ):
            self._log_or_print(
                utime,
                "CHARGING",
                f"Starting battery charge at RA={command.ra:.2f} Dec={command.dec:.2f} obsid={command.obsid}",
            )
            self._enqueue_slew(
                command.ra, command.dec, command.obsid, utime, obstype="CHARGE"
            )

    def _end_battery_charge(self, utime: float) -> None:
        """Handle END_BATTERY_CHARGE command execution.

        Terminates charging mode by returning to previous science pointing.
        """
        self._log_or_print(utime, "CHARGING", "Ending battery charge")

        # Clear the charging slew state immediately so _is_in_charging_mode returns False
        # This prevents staying in CHARGING mode while slewing back to science
        if self.last_slew is not None and self.last_slew.obstype == "CHARGE":
            self.last_slew = None

        # Return to the previous science PPT if one exists
        if self.last_ppt is not None:
            self._log_or_print(
                utime,
                "CHARGING",
                f"Returning to last PPT at RA={self.last_ppt.endra:.2f} Dec={self.last_ppt.enddec:.2f} obsid={self.last_ppt.obsid}",
            )
            self._enqueue_slew(
                self.last_ppt.endra,
                self.last_ppt.enddec,
                self.last_ppt.obsid,
                utime,
            )



"""Tests for communications system configuration."""

import pytest

from conops.common.enums import AntennaType, Polarization
from conops.config import (
    AntennaPointing,
    BandCapability,
    CommunicationsSystem,
)


class TestBandCapability:
    """Test BandCapability model."""

    def test_band_capability_creation(self):
        """Test creating a band capability."""
        band = BandCapability(band="X", uplink_rate_mbps=5.0, downlink_rate_mbps=50.0)
        assert band.band == "X"
        assert band.uplink_rate_mbps == 5.0
        assert band.downlink_rate_mbps == 50.0

    def test_band_capability_defaults(self):
        """Test default values for band capability."""
        band = BandCapability(band="S")
        assert band.band == "S"
        # Defaults now auto-populate for known bands
        assert band.uplink_rate_mbps == 2.0
        assert band.downlink_rate_mbps == 10.0

    def test_band_capability_validation(self):
        """Test validation of band capability values."""
        # Should not allow negative rates
        with pytest.raises(Exception):  # pydantic validation error
            BandCapability(band="X", uplink_rate_mbps=-1.0)

        with pytest.raises(Exception):  # pydantic validation error
            BandCapability(band="X", downlink_rate_mbps=-1.0)

    def test_supported_bands(self):
        """Test all supported band types."""
        for band_type in ["S", "X", "Ka", "Ku", "L", "C", "K"]:
            band = BandCapability(band=band_type, downlink_rate_mbps=10.0)
            assert band.band == band_type


class TestAntennaPointing:
    """Test AntennaPointing model."""

    def test_omni_antenna(self):
        """Test omnidirectional antenna configuration."""
        pointing = AntennaPointing(antenna_type=AntennaType.OMNI)
        assert pointing.antenna_type == AntennaType.OMNI
        assert not pointing.is_nadir_pointing()

    def test_fixed_nadir_pointing(self):
        """Test fixed nadir-pointing antenna (default)."""
        pointing = AntennaPointing(
            antenna_type=AntennaType.FIXED,
            fixed_azimuth_deg=0.0,
            fixed_elevation_deg=0.0,
        )
        assert pointing.antenna_type == AntennaType.FIXED
        assert pointing.is_nadir_pointing()

    def test_fixed_non_nadir_pointing(self):
        """Test fixed antenna not pointing nadir."""
        pointing = AntennaPointing(
            antenna_type=AntennaType.FIXED,
            fixed_azimuth_deg=180.0,
            fixed_elevation_deg=0.0,
        )
        assert pointing.antenna_type == AntennaType.FIXED
        assert not pointing.is_nadir_pointing()  # Zenith pointing

    def test_gimbaled_antenna(self):
        """Test gimbaled antenna configuration."""
        pointing = AntennaPointing(
            antenna_type=AntennaType.GIMBALED, gimbal_range_deg=30.0
        )
        assert pointing.antenna_type == AntennaType.GIMBALED
        assert pointing.gimbal_range_deg == 30.0
        assert not pointing.is_nadir_pointing()

    def test_antenna_pointing_validation(self):
        """Test validation of antenna pointing angles."""
        # Azimuth must be 0-360
        with pytest.raises(Exception):
            AntennaPointing(antenna_type=AntennaType.FIXED, fixed_azimuth_deg=400.0)

        # Elevation must be -90 to 90
        with pytest.raises(Exception):
            AntennaPointing(antenna_type=AntennaType.FIXED, fixed_elevation_deg=100.0)

        # Gimbal range must be 0-180
        with pytest.raises(Exception):
            AntennaPointing(antenna_type=AntennaType.GIMBALED, gimbal_range_deg=200.0)


class TestPolarization:
    """Test Polarization enum."""

    def test_all_polarizations(self):
        """Test all polarization types are accessible."""
        assert Polarization.LINEAR_HORIZONTAL == "linear_horizontal"
        assert Polarization.LINEAR_VERTICAL == "linear_vertical"
        assert Polarization.CIRCULAR_RIGHT == "circular_right"
        assert Polarization.CIRCULAR_LEFT == "circular_left"
        assert Polarization.DUAL == "dual"


class TestCommunicationsSystem:
    """Test CommunicationsSystem model."""

    def test_default_communications_system(self):
        """Test default communications system."""
        comms = CommunicationsSystem()
        assert comms.name == "Default Comms"
        assert len(comms.band_capabilities) == 1
        assert comms.band_capabilities[0].band == "S"
        assert comms.antenna_pointing.antenna_type == AntennaType.FIXED
        assert comms.polarization == Polarization.CIRCULAR_RIGHT
        assert comms.pointing_accuracy_deg == 5.0

    def test_custom_communications_system(self):
        """Test creating a custom communications system."""
        comms = CommunicationsSystem(
            name="X-band Comms",
            band_capabilities=[
                BandCapability(
                    band="X", uplink_rate_mbps=10.0, downlink_rate_mbps=100.0
                )
            ],
            antenna_pointing=AntennaPointing(
                antenna_type=AntennaType.GIMBALED, gimbal_range_deg=45.0
            ),
            polarization=Polarization.CIRCULAR_LEFT,
            pointing_accuracy_deg=2.0,
        )
        assert comms.name == "X-band Comms"
        assert len(comms.band_capabilities) == 1
        assert comms.band_capabilities[0].band == "X"
        assert comms.antenna_pointing.antenna_type == AntennaType.GIMBALED
        assert comms.antenna_pointing.gimbal_range_deg == 45.0
        assert comms.polarization == Polarization.CIRCULAR_LEFT
        assert comms.pointing_accuracy_deg == 2.0

    def test_multi_band_system(self):
        """Test communications system with multiple bands."""
        comms = CommunicationsSystem(
            name="Multi-band Comms",
            band_capabilities=[
                BandCapability(band="S", uplink_rate_mbps=2.0, downlink_rate_mbps=10.0),
                BandCapability(
                    band="X", uplink_rate_mbps=10.0, downlink_rate_mbps=100.0
                ),
                BandCapability(
                    band="Ka", uplink_rate_mbps=20.0, downlink_rate_mbps=500.0
                ),
            ],
        )
        assert len(comms.band_capabilities) == 3
        assert comms.band_capabilities[0].band == "S"
        assert comms.band_capabilities[1].band == "X"
        assert comms.band_capabilities[2].band == "Ka"

    def test_get_band(self):
        """Test retrieving a specific band capability."""
        comms = CommunicationsSystem(
            band_capabilities=[
                BandCapability(band="S", downlink_rate_mbps=10.0),
                BandCapability(band="X", downlink_rate_mbps=100.0),
            ]
        )
        s_band = comms.get_band("S")
        assert s_band is not None
        assert s_band.band == "S"
        assert s_band.downlink_rate_mbps == 10.0

        x_band = comms.get_band("X")
        assert x_band is not None
        assert x_band.band == "X"
        assert x_band.downlink_rate_mbps == 100.0

        # Non-existent band
        ka_band = comms.get_band("Ka")
        assert ka_band is None

    def test_get_downlink_rate(self):
        """Test getting downlink rate for a band."""
        comms = CommunicationsSystem(
            band_capabilities=[
                BandCapability(band="X", downlink_rate_mbps=150.0),
            ]
        )
        assert comms.get_downlink_rate("X") == 150.0
        assert comms.get_downlink_rate("S") == 0.0  # Not supported

    def test_get_uplink_rate(self):
        """Test getting uplink rate for a band."""
        comms = CommunicationsSystem(
            band_capabilities=[
                BandCapability(band="X", uplink_rate_mbps=15.0),
            ]
        )
        assert comms.get_uplink_rate("X") == 15.0
        assert comms.get_uplink_rate("Ka") == 0.0  # Not supported

    def test_can_communicate_omni(self):
        """Test communication check with omnidirectional antenna."""
        comms = CommunicationsSystem(
            antenna_pointing=AntennaPointing(antenna_type=AntennaType.OMNI),
            pointing_accuracy_deg=5.0,
        )
        # Omni antenna should always be able to communicate
        assert comms.can_communicate(0.0)
        assert comms.can_communicate(10.0)
        assert comms.can_communicate(90.0)
        assert comms.can_communicate(180.0)

    def test_can_communicate_fixed(self):
        """Test communication check with fixed antenna."""
        comms = CommunicationsSystem(
            antenna_pointing=AntennaPointing(antenna_type=AntennaType.FIXED),
            pointing_accuracy_deg=5.0,
        )
        # Within pointing accuracy
        assert comms.can_communicate(0.0)
        assert comms.can_communicate(3.0)
        assert comms.can_communicate(5.0)

        # Outside pointing accuracy
        assert not comms.can_communicate(6.0)
        assert not comms.can_communicate(10.0)
        assert not comms.can_communicate(90.0)

    def test_can_communicate_gimbaled(self):
        """Test communication check with gimbaled antenna."""
        comms = CommunicationsSystem(
            antenna_pointing=AntennaPointing(
                antenna_type=AntennaType.GIMBALED, gimbal_range_deg=30.0
            ),
            pointing_accuracy_deg=10.0,
        )
        # Within pointing accuracy
        assert comms.can_communicate(5.0)
        assert comms.can_communicate(10.0)

        # Outside pointing accuracy
        assert not comms.can_communicate(11.0)
        assert not comms.can_communicate(20.0)

    def test_pointing_accuracy_validation(self):
        """Test pointing accuracy validation."""
        # Valid range
        comms = CommunicationsSystem(pointing_accuracy_deg=1.0)
        assert comms.pointing_accuracy_deg == 1.0

        # Should not allow negative
        with pytest.raises(Exception):
            CommunicationsSystem(pointing_accuracy_deg=-1.0)

        # Should not allow > 180
        with pytest.raises(Exception):
            CommunicationsSystem(pointing_accuracy_deg=200.0)


class TestCommunicationsIntegration:
    """Integration tests for communications system."""

    def test_complete_x_band_system(self):
        """Test a complete X-band communications system configuration."""
        comms = CommunicationsSystem(
            name="High-Rate X-band",
            band_capabilities=[
                BandCapability(
                    band="X", uplink_rate_mbps=20.0, downlink_rate_mbps=300.0
                )
            ],
            antenna_pointing=AntennaPointing(
                antenna_type=AntennaType.GIMBALED, gimbal_range_deg=60.0
            ),
            polarization=Polarization.CIRCULAR_RIGHT,
            pointing_accuracy_deg=3.0,
        )

        # Verify configuration
        assert comms.name == "High-Rate X-band"
        assert comms.get_downlink_rate("X") == 300.0
        assert comms.get_uplink_rate("X") == 20.0
        assert comms.antenna_pointing.antenna_type == AntennaType.GIMBALED
        assert comms.can_communicate(2.0)
        assert not comms.can_communicate(5.0)

    def test_nadir_pointing_s_band(self):
        """Test a nadir-pointing S-band system (typical for LEO)."""
        comms = CommunicationsSystem(
            name="LEO S-band",
            band_capabilities=[
                BandCapability(band="S", uplink_rate_mbps=1.0, downlink_rate_mbps=5.0)
            ],
            antenna_pointing=AntennaPointing(
                antenna_type=AntennaType.FIXED,
                fixed_azimuth_deg=0.0,
                fixed_elevation_deg=0.0,
            ),
            polarization=Polarization.CIRCULAR_RIGHT,
            pointing_accuracy_deg=10.0,
        )

        assert comms.antenna_pointing.is_nadir_pointing()
        assert comms.get_downlink_rate("S") == 5.0
        assert comms.can_communicate(9.0)
        assert not comms.can_communicate(11.0)

    def test_dual_band_system(self):
        """Test a dual-band S/X communications system."""
        comms = CommunicationsSystem(
            name="S/X Dual Band",
            band_capabilities=[
                BandCapability(band="S", uplink_rate_mbps=1.0, downlink_rate_mbps=10.0),
                BandCapability(
                    band="X", uplink_rate_mbps=10.0, downlink_rate_mbps=150.0
                ),
            ],
            antenna_pointing=AntennaPointing(
                antenna_type=AntennaType.GIMBALED, gimbal_range_deg=45.0
            ),
            polarization=Polarization.DUAL,
            pointing_accuracy_deg=5.0,
        )

        # Both bands should be available
        assert comms.get_downlink_rate("S") == 10.0
        assert comms.get_downlink_rate("X") == 150.0
        assert comms.get_uplink_rate("S") == 1.0
        assert comms.get_uplink_rate("X") == 10.0
        assert comms.polarization == Polarization.DUAL



"""Test fixtures for slew subsystem tests."""

from unittest.mock import Mock

import numpy as np
import pytest

from conops import Slew


@pytest.fixture
def ephem():
    return Mock()


@pytest.fixture
def constraint(ephem):
    constraint = Mock()
    constraint.ephem = ephem
    return constraint


@pytest.fixture
def acs_config():
    return Mock()


@pytest.fixture
def mock_config(constraint, acs_config):
    config = Mock()
    config.constraint = constraint
    config.spacecraft_bus = Mock()
    config.spacecraft_bus.attitude_control = acs_config
    return config


@pytest.fixture
def slew(mock_config):
    return Slew(config=mock_config)


@pytest.fixture
def slew_with_positions(slew):
    slew.startra = 45.0
    slew.startdec = 30.0
    slew.endra = 90.0
    slew.enddec = 60.0
    slew.slewstart = 1700000000.0
    return slew


@pytest.fixture
def slew_slewing(slew):
    slew.slewstart = 1700000000.0
    slew.slewend = 1700000100.0
    return slew


@pytest.fixture
def slew_ra_dec(slew):
    slew.startra = 45.0
    slew.startdec = 30.0
    slew.endra = 90.0
    slew.enddec = 60.0
    slew.slewstart = 1700000000.0
    slew.slewend = 1700000100.0
    slew.slewpath = (np.array([45.0, 90.0]), np.array([30.0, 60.0]))
    slew.slewsecs = np.array([0.0, 100.0])
    return slew


@pytest.fixture
def slew_setup(slew, acs_config):
    acs_config.s_of_t = Mock(return_value=0.0)
    acs_config.motion_time = Mock(return_value=100.0)
    slew.startra = 45.0
    slew.startdec = 30.0
    slew.endra = 90.0
    slew.enddec = 60.0
    slew.slewstart = 1700000000.0
    slew.slewend = 1700000100.0
    slew.slewpath = (np.array([45.0, 90.0]), np.array([30.0, 60.0]))
    slew.slewdist = 50.0
    return slew


@pytest.fixture
def slew_interpolation(slew, acs_config):
    acs_config.s_of_t = Mock(return_value=25.0)
    acs_config.motion_time = Mock(return_value=100.0)
    slew.startra = 45.0
    slew.startdec = 30.0
    slew.endra = 90.0
    slew.enddec = 60.0
    slew.slewstart = 1700000000.0
    slew.slewend = 1700000100.0
    slew.slewpath = (np.array([45.0, 90.0]), np.array([30.0, 60.0]))
    slew.slewdist = 50.0
    return slew


@pytest.fixture
def slew_modern_path(slew, acs_config):
    acs_config.s_of_t = Mock(return_value=0.0)
    acs_config.motion_time = Mock(return_value=100.0)
    slew.startra = 45.0
    slew.startdec = 30.0
    slew.endra = 90.0
    slew.enddec = 60.0
    slew.slewstart = 1700000000.0
    slew.slewend = 1700000100.0
    slew.slewpath = (np.array([67.5]), np.array([45.0]))
    slew.slewdist = 50.0
    return slew


@pytest.fixture
def slew_acs(slew, acs_config):
    acs_config.motion_time = Mock(return_value=100.0)
    acs_config.s_of_t = Mock(return_value=50.0)
    slew.startra = 0.0
    slew.startdec = 0.0
    slew.endra = 10.0
    slew.enddec = 10.0
    slew.slewstart = 1700000000.0
    slew.slewend = 1700000100.0
    slew.slewdist = 14.142
    slew.slewpath = (np.linspace(0.0, 10.0, 20), np.linspace(0.0, 10.0, 20))
    return slew


@pytest.fixture
def slew_interp_start(slew):
    slew.startra = 0.0
    slew.startdec = 0.0
    slew.slewstart = 1700000000.0
    slew.slewpath = (np.array([0.0, 10.0]), np.array([0.0, 10.0]))
    slew.slewsecs = np.array([0.0, 100.0])
    slew.slewdist = 0
    return slew


@pytest.fixture
def slew_calc_setup(slew, acs_config):
    acs_config.predict_slew = Mock(
        return_value=(10.0, (np.array([0.0, 10.0]), np.array([0.0, 10.0])))
    )
    acs_config.slew_time = Mock(return_value=50.0)
    slew.startra = 0.0
    slew.startdec = 0.0
    slew.endra = 10.0
    slew.enddec = 10.0
    slew.slewstart = 1700000000.0
    return slew


@pytest.fixture
def slew_calc_setup_alt(slew, acs_config):
    acs_config.predict_slew = Mock(
        return_value=(5.0, (np.array([0.0, 5.0]), np.array([0.0, 5.0])))
    )
    acs_config.slew_time = Mock(return_value=30.0)
    slew.startra = 0.0
    slew.startdec = 0.0
    slew.endra = 5.0
    slew.enddec = 5.0
    slew.slewstart = 1700000000.0
    return slew


@pytest.fixture
def slew_predict_setup(slew, acs_config):
    ra_path = np.linspace(45.0, 90.0, 20)
    dec_path = np.linspace(30.0, 60.0, 20)
    acs_config.predict_slew = Mock(return_value=(14.142, (ra_path, dec_path)))
    slew.startra = 45.0
    slew.startdec = 30.0
    slew.endra = 90.0
    slew.enddec = 60.0
    return slew, ra_path, dec_path



from __future__ import annotations

from collections.abc import Iterator

from .plan_entry import PlanEntry


class TargetList:
    """List of potential targets for Spacecraft to observe"""

    targets: list[PlanEntry]

    def __init__(self) -> None:
        self.targets = list()

    def __getitem__(self, number: int) -> PlanEntry:
        return self.targets[number]

    def __iter__(self) -> Iterator[PlanEntry]:
        return iter(self.targets)

    def add_target(self, plan_entry: PlanEntry) -> None:
        self.targets.append(plan_entry)

    def __len__(self) -> int:
        return len(self.targets)


class Plan:
    """Simple Plan class"""

    entries: list[PlanEntry]

    def __init__(self) -> None:
        self.entries = list()

    def __getitem__(self, number: int) -> PlanEntry:
        return self.entries[number]

    def __len__(self) -> int:
        return len(self.entries)

    def __iter__(self) -> Iterator[PlanEntry]:
        return iter(self.entries)

    def which_ppt(self, utime: float) -> PlanEntry | None:
        """At a given utime, which PPT is the current one?"""
        for ppt in self.entries:
            if ppt.begin <= utime < ppt.end:
                return ppt
        return None

    def extend(self, ppt: list[PlanEntry]) -> None:
        self.entries.extend(ppt)

    def append(self, ppt: PlanEntry) -> None:
        self.entries.append(ppt)



from typing import Literal

import numpy as np

from ..common import unixtime2date
from ..config import MissionConfig
from .plan_entry import PlanEntry


class Pointing(PlanEntry):
    """Define the basic parameters of an observing target with visibility checking."""

    ra: float
    dec: float
    obsid: int
    name: str
    merit: float
    isat: bool

    def __init__(
        self,
        config: MissionConfig | None = None,
        ra: float = 0.0,
        dec: float = 0.0,
        obsid: int = 0,
        name: str = "FakeTarget",
        merit: float = 100.0,
        exptime: int = 1000,
        ss_min: int = 300,
        ss_max: int = 86400,
    ):
        # Handle both old and new parameter styles for backward compatibility
        if config is None:
            raise ValueError("Config must be provided to Pointing")

        PlanEntry.__init__(self, config=config, exptime=exptime)
        assert config.constraint is not None, "Constraint not properly set in Pointing"
        self.done = False
        self.obstype = "AT"
        self.isat = False
        self.ra = ra
        self.dec = dec
        self.targetid = obsid
        self.obsid = obsid
        self.name = name
        # ``fom`` is maintained as a legacy alias for ``merit`` for
        # backwards compatibility (e.g. tests and older code). The
        # canonical field we use internally is ``merit`` which can be
        # recomputed each scheduling iteration by ``Queue.meritsort``.
        self.fom = merit
        self.merit = merit
        self._done = False
        # Snapshot min/max size
        self.ss_min = ss_min  # seconds
        self.ss_max = ss_max  # seconds

    def in_sun(self, utime: float) -> bool:
        """Is this target in Sun constraint?"""
        in_sun = self.config.constraint.in_sun(self.ra, self.dec, utime)
        return in_sun

    def in_earth(self, utime: float) -> bool:
        """Is this target in Earth constraint?"""
        return self.config.constraint.in_earth(self.ra, self.dec, utime)

    def in_moon(self, utime: float) -> bool:
        """Is this target in Moon constraint?"""
        return self.config.constraint.in_moon(self.ra, self.dec, utime)

    def in_panel(self, utime: float) -> bool:
        """Is this target in Panel constraint?"""
        return self.config.constraint.in_panel(self.ra, self.dec, utime)

    def next_vis(self, utime: float) -> float | Literal[False]:
        """When is this target visible next?"""
        # Are we currently in a visibility window, if yes, return back the current time
        if self.visible(utime, utime):
            return utime

        # Are there no visibility windows? Then just return False
        if len(self.windows) == 0:
            return False
        try:
            visstarts = np.array(self.windows).transpose()[0]
            windex = np.where(visstarts - utime > 0)[0][0]
            return float(visstarts[windex])
        except Exception:
            return False

    def __str__(self) -> str:
        return f"{unixtime2date(self.begin)} {self.name} ({self.targetid}) RA={self.ra:.4f}, Dec={self.dec:4f}, Roll={self.roll:.1f}, Merit={self.merit}"

    @property
    def done(self) -> bool:
        if self.exptime is not None and self.exptime <= 0:
            self._done = True
        return self._done

    @done.setter
    def done(self, v: bool) -> None:
        self._done = v

    def reset(self) -> None:
        if self._exporig is not None:
            self._exptime = self._exporig
        self.done = False
        self.begin = 0
        self.end = 0
        self.slewtime = 0



"""Additional tests to achieve 100% coverage for ACS class."""

from unittest.mock import Mock, patch

from conops import ACSCommand, ACSCommandType, ACSMode, Pass, Slew


class TestExecuteCommandCoverage:
    """Test command execution handler methods."""

    def test_end_pass_adds_slew_with_last_ppt(self, acs):
        """END_PASS should NOT call enqueue_command for last_ppt in queue-driven mode."""
        mock_ppt = Mock(spec=Slew)
        mock_ppt.endra = 45.0
        mock_ppt.enddec = 30.0
        mock_ppt.obsid = 100
        acs.last_ppt = mock_ppt
        acs.current_pass = Mock(spec=Pass)
        acs.acsmode = ACSMode.PASS

        # Directly call _end_pass
        acs._end_pass(1514764800.0)
        # Verify currentpass is cleared and mode is set to SCIENCE
        assert acs.current_pass is None
        assert acs.acsmode == ACSMode.SCIENCE

    def test_end_pass_clears_currentpass(self, acs):
        mock_ppt = Mock(spec=Slew)
        mock_ppt.endra = 45.0
        mock_ppt.enddec = 30.0
        mock_ppt.obsid = 100
        acs.last_ppt = mock_ppt
        acs.current_pass = Mock(spec=Pass)
        acs.acsmode = ACSMode.PASS

        acs._end_pass(1514764800.0)
        assert acs.current_pass is None
        assert acs.acsmode == ACSMode.SCIENCE

    def test_end_pass_sets_mode_science(self, acs):
        mock_ppt = Mock(spec=Slew)
        mock_ppt.endra = 45.0
        mock_ppt.enddec = 30.0
        mock_ppt.obsid = 100
        acs.last_ppt = mock_ppt
        acs.current_pass = Mock(spec=Pass)
        acs.acsmode = ACSMode.PASS

        acs._end_pass(1514764800.0)
        assert acs.acsmode == ACSMode.SCIENCE

    def test_end_pass_no_last_ppt_clears_currentpass(self, acs):
        acs.last_ppt = None
        acs.current_pass = Mock(spec=Pass)
        acs.acsmode = ACSMode.PASS

        acs._end_pass(1514764800.0)
        assert acs.current_pass is None

    def test_end_pass_no_last_ppt_sets_mode_science(self, acs):
        acs.last_ppt = None
        acs.current_pass = Mock(spec=Pass)
        acs.acsmode = ACSMode.PASS

        acs._end_pass(1514764800.0)
        assert acs.acsmode == ACSMode.SCIENCE

    def test_execute_null_slew_does_not_start(self, acs):
        command = ACSCommand(
            command_type=ACSCommandType.SLEW_TO_TARGET,
            execution_time=1514764800.0,
            slew=None,
        )

        with patch.object(acs, "_start_slew") as mock_start_slew:
            acs._handle_slew_command(command, 1514764800.0)
            mock_start_slew.assert_not_called()

    def test_execute_slew_to_target_none_slew_does_not_start(self, acs):
        command = ACSCommand(
            command_type=ACSCommandType.SLEW_TO_TARGET,
            execution_time=1514764800.0,
            slew=None,
        )

        with patch.object(acs, "_start_slew") as mock_start_slew:
            acs._handle_slew_command(command, 1514764800.0)
            mock_start_slew.assert_not_called()

    def test_execute_start_pass_none_slew_does_not_start(self, acs):
        command = ACSCommand(
            command_type=ACSCommandType.START_PASS,
            execution_time=1514764800.0,
            slew=None,
        )

        with patch.object(acs, "_start_slew") as mock_start_slew:
            acs._start_pass(command, 1514764800.0)
            mock_start_slew.assert_not_called()

    def test_execute_start_pass_non_pass_slew_does_not_start(self, acs):
        mock_slew = Mock(spec=Slew)
        command = ACSCommand(
            command_type=ACSCommandType.START_PASS,
            execution_time=1514764800.0,
            slew=mock_slew,
        )

        with patch.object(acs, "_start_slew") as mock_start_slew:
            acs._start_pass(command, 1514764800.0)
            mock_start_slew.assert_not_called()


class TestStartSlewCoverage:
    """Test _start_slew behavior - ACS always drives spacecraft from current position."""

    def test_start_slew_sets_startra_from_acs(self, acs):
        """Slew always starts from current ACS pointing."""
        acs.ra = 10.0
        acs.dec = 20.0

        mock_slew = Mock(spec=Slew)
        mock_slew.startra = 0.0
        mock_slew.startdec = 0.0
        mock_slew.endra = 45.0
        mock_slew.enddec = 30.0
        mock_slew.obstype = "PPT"
        mock_slew.slewstart = 1514764800.0
        mock_slew.slewtime = 60.0
        mock_slew.calc_slewtime = Mock()

        acs._start_slew(mock_slew, 1514764800.0)
        assert mock_slew.startra == 10.0

    def test_start_slew_sets_startdec_from_acs(self, acs):
        """Slew always starts from current ACS pointing."""
        acs.ra = 10.0
        acs.dec = 20.0

        mock_slew = Mock(spec=Slew)
        mock_slew.startra = 0.0
        mock_slew.startdec = 0.0
        mock_slew.endra = 45.0
        mock_slew.enddec = 30.0
        mock_slew.obstype = "PPT"
        mock_slew.slewstart = 1514764800.0
        mock_slew.slewtime = 60.0
        mock_slew.calc_slewtime = Mock()

        acs._start_slew(mock_slew, 1514764800.0)
        assert mock_slew.startdec == 20.0

    def test_start_slew_always_calls_calc_slewtime(self, acs):
        """calc_slewtime is always called to compute the new slew profile."""
        acs.ra = 10.0
        acs.dec = 20.0

        mock_slew = Mock(spec=Slew)
        mock_slew.startra = 0.0
        mock_slew.startdec = 0.0
        mock_slew.endra = 45.0
        mock_slew.enddec = 30.0
        mock_slew.obstype = "PPT"
        mock_slew.slewstart = 1514764800.0
        mock_slew.slewtime = 60.0
        mock_slew.calc_slewtime = Mock()

        acs._start_slew(mock_slew, 1514764800.0)
        mock_slew.calc_slewtime.assert_called_once()

    def test_start_slew_sets_pass_startra_from_acs(self, acs):
        """Pass slews also start from current ACS pointing."""
        acs.ra = 10.0
        acs.dec = 20.0

        mock_pass = Mock(spec=Pass)
        mock_pass.startra = 0.0
        mock_pass.startdec = 0.0
        mock_pass.endra = 45.0
        mock_pass.enddec = 30.0
        mock_pass.obstype = "GSP"
        mock_pass.slewstart = 1514764800.0
        mock_pass.slewtime = 60.0
        mock_pass.calc_slewtime = Mock()

        acs._start_slew(mock_pass, 1514764800.0)
        assert mock_pass.startra == 10.0

    def test_start_slew_sets_pass_startdec_from_acs(self, acs):
        """Pass slews also start from current ACS pointing."""
        acs.ra = 10.0
        acs.dec = 20.0

        mock_pass = Mock(spec=Pass)
        mock_pass.startra = 0.0
        mock_pass.startdec = 0.0
        mock_pass.endra = 45.0
        mock_pass.enddec = 30.0
        mock_pass.obstype = "GSP"
        mock_pass.slewstart = 1514764800.0
        mock_pass.slewtime = 60.0
        mock_pass.calc_slewtime = Mock()

        acs._start_slew(mock_pass, 1514764800.0)
        assert mock_pass.startdec == 20.0

    def test_start_slew_calls_calc_for_pass(self, acs):
        """Pass slews also recalculate slew time."""
        acs.ra = 10.0
        acs.dec = 20.0

        mock_pass = Mock(spec=Pass)
        mock_pass.startra = 0.0
        mock_pass.startdec = 0.0
        mock_pass.endra = 45.0
        mock_pass.enddec = 30.0
        mock_pass.obstype = "GSP"
        mock_pass.slewstart = 1514764800.0
        mock_pass.slewtime = 60.0
        mock_pass.calc_slewtime = Mock()

        acs._start_slew(mock_pass, 1514764800.0)
        mock_pass.calc_slewtime.assert_called_once()

    def test_start_slew_sets_start_from_zero_position(self, acs):
        """Even when ACS is at origin, slew starts from there."""
        acs.ra = 0.0
        acs.dec = 0.0

        mock_slew = Mock(spec=Slew)
        mock_slew.startra = 5.0
        mock_slew.startdec = 10.0
        mock_slew.endra = 45.0
        mock_slew.enddec = 30.0
        mock_slew.obstype = "PPT"
        mock_slew.slewstart = 1514764800.0
        mock_slew.slewtime = 60.0
        mock_slew.calc_slewtime = Mock()

        acs._start_slew(mock_slew, 1514764800.0)
        assert mock_slew.startra == 0.0
        assert mock_slew.startdec == 0.0

    def test_start_slew_sets_slewstart_to_current_time(self, acs):
        """Slew start time is set to execution time."""
        acs.ra = 10.0
        acs.dec = 20.0

        mock_slew = Mock(spec=Slew)
        mock_slew.startra = 10.0
        mock_slew.startdec = 20.0
        mock_slew.endra = 45.0
        mock_slew.enddec = 30.0
        mock_slew.obstype = "PPT"
        mock_slew.slewstart = 9999999.0  # Different from execution time
        mock_slew.slewtime = 60.0
        mock_slew.calc_slewtime = Mock()

        acs._start_slew(mock_slew, 1514764800.0)
        assert mock_slew.slewstart == 1514764800.0

    def test_start_slew_overwrites_matching_startra(self, acs):
        """Even if startra already matches, it gets set (no special case)."""
        acs.ra = 10.0
        acs.dec = 20.0

        mock_slew = Mock(spec=Slew)
        mock_slew.startra = 10.0  # Already matches
        mock_slew.startdec = 20.0
        mock_slew.endra = 45.0
        mock_slew.enddec = 30.0
        mock_slew.obstype = "PPT"
        mock_slew.slewstart = 1514764800.0
        mock_slew.slewtime = 60.0
        mock_slew.calc_slewtime = Mock()

        acs._start_slew(mock_slew, 1514764800.0)
        # startra is set to acs.ra (same value)
        assert mock_slew.startra == 10.0
        # calc_slewtime is still called
        mock_slew.calc_slewtime.assert_called_once()

    def test_start_slew_updates_last_ppt(self, acs):
        acs.ra = 10.0
        acs.dec = 20.0

        mock_slew = Mock(spec=Slew)
        mock_slew.startra = 10.0
        mock_slew.startdec = 20.0
        mock_slew.endra = 45.0
        mock_slew.enddec = 30.0
        mock_slew.obstype = "PPT"
        mock_slew.slewstart = 1514764800.0
        mock_slew.slewtime = 60.0
        mock_slew.calc_slewtime = Mock()

        acs._start_slew(mock_slew, 1514764800.0)
        assert acs.last_ppt == mock_slew

    def test_start_slew_no_last_ppt_for_non_ppt(self, acs):
        acs.ra = 10.0
        acs.dec = 20.0

        mock_slew = Mock(spec=Slew)
        mock_slew.startra = 10.0
        mock_slew.startdec = 20.0
        mock_slew.endra = 45.0
        mock_slew.enddec = 30.0
        mock_slew.obstype = "GSP"
        mock_slew.slewstart = 1514764800.0
        mock_slew.slewtime = 60.0
        mock_slew.calc_slewtime = Mock()

        acs.last_ppt = None

        acs._start_slew(mock_slew, 1514764800.0)
        assert acs.last_ppt is None

        acs._start_slew(mock_slew, 1514764800.0)
        assert acs.last_ppt is None


class TestEndPassCoverage:
    """Test _end_pass method."""

    def test_end_pass_adds_slew_returning_to_last_ppt(self, acs):
        mock_ppt = Mock(spec=Slew)
        mock_ppt.endra = 45.0
        mock_ppt.enddec = 30.0
        mock_ppt.obsid = 100
        acs.last_ppt = mock_ppt

        acs.current_pass = Mock(spec=Pass)
        acs.acsmode = ACSMode.PASS

        with patch.object(
            acs, "enqueue_command", return_value=True
        ) as mock_enqueue_command:
            acs._end_pass(1514764800.0)
            mock_enqueue_command.assert_not_called()

    def test_end_pass_clears_currentpass(self, acs):
        mock_ppt = Mock(spec=Slew)
        mock_ppt.endra = 45.0
        mock_ppt.enddec = 30.0
        mock_ppt.obsid = 100
        acs.last_ppt = mock_ppt

        acs.current_pass = Mock(spec=Pass)
        acs.acsmode = ACSMode.PASS

        with patch.object(acs, "enqueue_command", return_value=True):
            acs._end_pass(1514764800.0)
            assert acs.current_pass is None

    def test_end_pass_sets_mode_science_on_end(self, acs):
        mock_ppt = Mock(spec=Slew)
        mock_ppt.endra = 45.0
        mock_ppt.enddec = 30.0
        mock_ppt.obsid = 100
        acs.last_ppt = mock_ppt

        acs.current_pass = Mock(spec=Pass)
        acs.acsmode = ACSMode.PASS

        with patch.object(acs, "enqueue_command", return_value=True):
            acs._end_pass(1514764800.0)
            assert acs.acsmode == ACSMode.SCIENCE

    def test_end_pass_with_no_last_ppt_clears_currentpass(self, acs):
        acs.last_ppt = None
        acs.current_pass = Mock(spec=Pass)
        acs.acsmode = ACSMode.PASS

        acs._end_pass(1514764800.0)
        assert acs.current_pass is None

    def test_end_pass_with_no_last_ppt_sets_mode_science(self, acs):
        acs.last_ppt = None
        acs.current_pass = Mock(spec=Pass)
        acs.acsmode = ACSMode.PASS

        acs._end_pass(1514764800.0)
        assert acs.acsmode == ACSMode.SCIENCE

    def test_end_pass_no_last_ppt_does_not_enqueue_slew(self, acs):
        acs.last_ppt = None
        acs.current_pass = Mock(spec=Pass)
        acs.acsmode = ACSMode.PASS

        acs._end_pass(1514764800.0)
        assert len(acs.command_queue) == 0


class TestProcessCommandsCoverage:
    """Test _process_commands to ensure queue processing is covered."""

    def _make_mock_slew(
        self, startra=0.0, startdec=0.0, endra=45.0, enddec=30.0, obstype="PPT"
    ):
        """Helper to create a mock slew with all required attributes."""
        mock_slew = Mock(spec=Slew)
        mock_slew.startra = startra
        mock_slew.startdec = startdec
        mock_slew.endra = endra
        mock_slew.enddec = enddec
        mock_slew.obstype = obstype
        mock_slew.slewstart = 1514764800.0
        mock_slew.slewtime = 60.0
        mock_slew.calc_slewtime = Mock()
        return mock_slew

    def test_process_commands_executes_first_due_command(self, acs):
        mock_slew1 = self._make_mock_slew(0.0, 0.0, 45.0, 30.0)
        mock_slew2 = self._make_mock_slew(45.0, 30.0, 90.0, 60.0)

        command1 = ACSCommand(
            command_type=ACSCommandType.SLEW_TO_TARGET,
            execution_time=1514764800.0,
            slew=mock_slew1,
        )
        command2 = ACSCommand(
            command_type=ACSCommandType.SLEW_TO_TARGET,
            execution_time=1514764810.0,
            slew=mock_slew2,
        )
        command3 = ACSCommand(
            command_type=ACSCommandType.SLEW_TO_TARGET,
            execution_time=1514764900.0,
            slew=mock_slew1,
        )
        acs.command_queue = [command1, command2, command3]

        acs._process_commands(1514764815.0)
        assert len(acs.executed_commands) >= 1

    def test_process_commands_executes_second_due_command(self, acs):
        mock_slew1 = self._make_mock_slew(0.0, 0.0, 45.0, 30.0)
        mock_slew2 = self._make_mock_slew(45.0, 30.0, 90.0, 60.0)

        command1 = ACSCommand(
            command_type=ACSCommandType.SLEW_TO_TARGET,
            execution_time=1514764800.0,
            slew=mock_slew1,
        )
        command2 = ACSCommand(
            command_type=ACSCommandType.SLEW_TO_TARGET,
            execution_time=1514764810.0,
            slew=mock_slew2,
        )
        command3 = ACSCommand(
            command_type=ACSCommandType.SLEW_TO_TARGET,
            execution_time=1514764900.0,
            slew=mock_slew1,
        )
        acs.command_queue = [command1, command2, command3]

        acs._process_commands(1514764815.0)
        assert len(acs.executed_commands) >= 2

    def test_process_commands_first_executed_command_matches_command1(self, acs):
        mock_slew1 = self._make_mock_slew(0.0, 0.0, 45.0, 30.0)
        mock_slew2 = self._make_mock_slew(45.0, 30.0, 90.0, 60.0)

        command1 = ACSCommand(
            command_type=ACSCommandType.SLEW_TO_TARGET,
            execution_time=1514764800.0,
            slew=mock_slew1,
        )
        command2 = ACSCommand(
            command_type=ACSCommandType.SLEW_TO_TARGET,
            execution_time=1514764810.0,
            slew=mock_slew2,
        )
        command3 = ACSCommand(
            command_type=ACSCommandType.SLEW_TO_TARGET,
            execution_time=1514764900.0,
            slew=mock_slew1,
        )
        acs.command_queue = [command1, command2, command3]

        acs._process_commands(1514764815.0)
        assert acs.executed_commands[0] == command1

    def test_process_commands_second_executed_command_matches_command2(self, acs):
        mock_slew1 = self._make_mock_slew(0.0, 0.0, 45.0, 30.0)
        mock_slew2 = self._make_mock_slew(45.0, 30.0, 90.0, 60.0)

        command1 = ACSCommand(
            command_type=ACSCommandType.SLEW_TO_TARGET,
            execution_time=1514764800.0,
            slew=mock_slew1,
        )
        command2 = ACSCommand(
            command_type=ACSCommandType.SLEW_TO_TARGET,
            execution_time=1514764810.0,
            slew=mock_slew2,
        )
        command3 = ACSCommand(
            command_type=ACSCommandType.SLEW_TO_TARGET,
            execution_time=1514764900.0,
            slew=mock_slew1,
        )
        acs.command_queue = [command1, command2, command3]

        acs._process_commands(1514764815.0)
        assert acs.executed_commands[1] == command2

    def test_process_commands_leaves_later_command_in_queue(self, acs):
        mock_slew1 = self._make_mock_slew(0.0, 0.0, 45.0, 30.0)
        mock_slew2 = self._make_mock_slew(45.0, 30.0, 90.0, 60.0)

        command1 = ACSCommand(
            command_type=ACSCommandType.SLEW_TO_TARGET,
            execution_time=1514764800.0,
            slew=mock_slew1,
        )
        command2 = ACSCommand(
            command_type=ACSCommandType.SLEW_TO_TARGET,
            execution_time=1514764810.0,
            slew=mock_slew2,
        )
        command3 = ACSCommand(
            command_type=ACSCommandType.SLEW_TO_TARGET,
            execution_time=1514764900.0,
            slew=mock_slew1,
        )
        acs.command_queue = [command1, command2, command3]

        acs._process_commands(1514764815.0)
        assert len(acs.command_queue) == 1

    def test_process_commands_remaining_queue_item_is_third(self, acs):
        mock_slew1 = self._make_mock_slew(0.0, 0.0, 45.0, 30.0)

        command1 = ACSCommand(
            command_type=ACSCommandType.SLEW_TO_TARGET,
            execution_time=1514764800.0,
            slew=mock_slew1,
        )
        command2 = ACSCommand(
            command_type=ACSCommandType.SLEW_TO_TARGET,
            execution_time=1514764810.0,
            slew=mock_slew1,
        )
        command3 = ACSCommand(
            command_type=ACSCommandType.SLEW_TO_TARGET,
            execution_time=1514764900.0,
            slew=mock_slew1,
        )
        acs.command_queue = [command1, command2, command3]

        acs._process_commands(1514764815.0)
        assert acs.command_queue[0] == command3


class TestExecuteCommandLogging:
    """Test command handler logging."""

    def test_handle_slew_command_executes(self, acs):
        mock_slew = Mock(spec=Slew)
        mock_slew.startra = 0.0
        mock_slew.startdec = 0.0
        mock_slew.endra = 45.0
        mock_slew.enddec = 30.0
        mock_slew.obstype = "PPT"
        mock_slew.slewstart = 1514764800.0
        mock_slew.slewtime = 60.0
        mock_slew.calc_slewtime = Mock()
        command1 = ACSCommand(
            command_type=ACSCommandType.SLEW_TO_TARGET,
            execution_time=1514764800.0,
            slew=mock_slew,
        )

        # Test that command executes without error
        acs._handle_slew_command(command1, 1514764800.0)
        assert acs.current_slew == mock_slew

    def test_start_pass_executes(self, acs):
        mock_slew = Mock(spec=Slew)
        mock_slew.startra = 0.0
        mock_slew.startdec = 0.0
        mock_slew.endra = 45.0
        mock_slew.enddec = 30.0
        mock_slew.obstype = "GSP"
        mock_slew.slewstart = 1514764800.0
        mock_slew.slewtime = 60.0
        mock_slew.calc_slewtime = Mock()
        command2 = ACSCommand(
            command_type=ACSCommandType.START_PASS,
            execution_time=1514764800.0,
            slew=mock_slew,
        )

        # Test that command executes without error
        acs._start_pass(command2, 1514764800.0)
        # _start_pass sets current_pass, not current_slew
        assert acs.acsmode == ACSMode.PASS


class TestGetModeCharging:
    """Test get_mode for CHARGING mode."""

    def test_get_mode_returns_slewing_when_in_eclipse(self, acs, monkeypatch):
        mock_slew = Mock(spec=Slew)
        mock_slew.obstype = "CHARGE"
        mock_slew.is_slewing = Mock(return_value=True)
        acs.current_slew = mock_slew

        mock_ephem = Mock()
        from datetime import datetime, timezone

        mock_ephem.timestamp = [datetime.fromtimestamp(1514764800.0, tz=timezone.utc)]
        mock_ephem.sun = [Mock()]
        mock_ephem.earth = [Mock()]
        mock_ephem.earth_radius_angle = [1.0]
        mock_ephem.in_eclipse = Mock(return_value=True)
        acs.ephem = mock_ephem

        # Mock constraint.in_eclipse to return True (in eclipse)
        monkeypatch.setattr(acs.constraint, "in_eclipse", lambda ra, dec, time: True)

        # Set ACS in_eclipse state to True
        acs.in_eclipse = True

        mode = acs.get_mode(1514764800.0)
        assert mode == ACSMode.SLEWING

    def test_get_mode_calls_in_eclipse_for_charging(self, acs, monkeypatch):
        mock_slew = Mock(spec=Slew)
        mock_slew.obstype = "CHARGE"
        mock_slew.is_slewing = Mock(return_value=True)
        acs.current_slew = mock_slew

        mock_ephem = Mock()
        from datetime import datetime, timezone

        mock_ephem.timestamp = [datetime.fromtimestamp(1514764800.0, tz=timezone.utc)]
        mock_ephem.sun = [Mock()]
        mock_ephem.earth = [Mock()]
        mock_ephem.earth_radius_angle = [1.0]
        mock_ephem.in_eclipse = Mock(return_value=True)
        acs.ephem = mock_ephem

        # Mock constraint.in_eclipse
        mock_in_eclipse = Mock(return_value=True)
        monkeypatch.setattr(acs.constraint, "in_eclipse", mock_in_eclipse)

        _ = acs.get_mode(1514764800.0)

    def test_get_mode_returns_charging_in_sunlight_while_slewing(
        self, acs, monkeypatch
    ):
        mock_slew = Mock(spec=Slew)
        mock_slew.obstype = "CHARGE"
        mock_slew.is_slewing = Mock(return_value=True)
        acs.current_slew = mock_slew

        mock_ephem = Mock()
        from datetime import datetime, timezone

        mock_ephem.timestamp = [datetime.fromtimestamp(1514764800.0, tz=timezone.utc)]
        mock_ephem.sun = [Mock()]
        mock_ephem.earth = [Mock()]
        mock_ephem.earth_radius_angle = [1.0]
        mock_ephem.in_eclipse = Mock(return_value=False)
        acs.ephem = mock_ephem

        # Mock constraint.in_eclipse to return False (in sunlight)
        monkeypatch.setattr(acs.constraint, "in_eclipse", lambda ra, dec, time: False)

        mode = acs.get_mode(1514764800.0)
        assert mode == ACSMode.CHARGING

    def test_get_mode_calls_in_eclipse_for_sunlight_slewing(self, acs, monkeypatch):
        mock_slew = Mock(spec=Slew)
        mock_slew.obstype = "CHARGE"
        mock_slew.is_slewing = Mock(return_value=True)
        acs.current_slew = mock_slew

        mock_ephem = Mock()
        from datetime import datetime, timezone

        mock_ephem.timestamp = [datetime.fromtimestamp(1514764800.0, tz=timezone.utc)]
        mock_ephem.sun = [Mock()]
        mock_ephem.earth = [Mock()]
        mock_ephem.earth_radius_angle = [1.0]
        mock_ephem.in_eclipse = Mock(return_value=False)
        acs.ephem = mock_ephem

        # Mock constraint.in_eclipse
        mock_in_eclipse = Mock(return_value=False)
        monkeypatch.setattr(acs.constraint, "in_eclipse", mock_in_eclipse)

        _ = acs.get_mode(1514764800.0)

    def test_get_mode_charging_in_dwell_when_not_slewing(self, acs, monkeypatch):
        mock_slew = Mock(spec=Slew)
        mock_slew.obstype = "CHARGE"
        mock_slew.is_slewing = Mock(return_value=False)
        acs.last_slew = mock_slew
        acs.current_slew = mock_slew

        mock_ephem = Mock()
        from datetime import datetime, timezone

        mock_ephem.timestamp = [datetime.fromtimestamp(1514764800.0, tz=timezone.utc)]
        mock_ephem.sun = [Mock()]
        mock_ephem.earth = [Mock()]
        mock_ephem.earth_radius_angle = [1.0]
        mock_ephem.in_eclipse = Mock(return_value=False)
        acs.ephem = mock_ephem

        # Mock constraint.in_eclipse to return False (in sunlight)
        monkeypatch.setattr(acs.constraint, "in_eclipse", lambda ra, dec, time: False)

        mode = acs.get_mode(1514764800.0)
        assert mode == ACSMode.CHARGING


class TestIsInChargingMode:
    """Test _is_in_charging_mode method."""

    def test_is_in_charging_mode_returns_true_when_ephem_lacks_in_eclipse(
        self, acs, monkeypatch
    ):
        mock_slew = Mock(spec=Slew)
        mock_slew.obstype = "CHARGE"
        mock_slew.is_slewing = Mock(return_value=False)
        acs.last_slew = mock_slew
        acs.current_slew = mock_slew

        mock_ephem = Mock(spec=["other_method"])
        from datetime import datetime, timezone

        mock_ephem.timestamp = [datetime.fromtimestamp(1514764800.0, tz=timezone.utc)]
        mock_ephem.sun = [Mock()]
        mock_ephem.earth = [Mock()]
        mock_ephem.earth_radius_angle = [1.0]
        acs.ephem = mock_ephem

        # Mock constraint.in_eclipse to return False (in sunlight)
        monkeypatch.setattr(acs.constraint, "in_eclipse", lambda ra, dec, time: False)

        assert acs._is_in_charging_mode(1514764800.0) is True


class TestBatteryChargingMethods:
    """Test battery charging request and execution methods."""

    def test_request_battery_charge_enqueues_command(self, acs):
        ra, dec, obsid = 45.0, 30.0, 0xBEEF
        utime = 1514764800.0

        acs.request_battery_charge(utime, ra, dec, obsid)
        assert len(acs.command_queue) == 1

    def test_request_battery_charge_creates_correct_command_type(self, acs):
        ra, dec, obsid = 45.0, 30.0, 0xBEEF
        utime = 1514764800.0

        acs.request_battery_charge(utime, ra, dec, obsid)
        cmd = acs.command_queue[0]
        assert cmd.command_type == ACSCommandType.START_BATTERY_CHARGE

    def test_request_battery_charge_sets_execution_time(self, acs):
        ra, dec, obsid = 45.0, 30.0, 0xBEEF
        utime = 1514764800.0

        acs.request_battery_charge(utime, ra, dec, obsid)
        cmd = acs.command_queue[0]
        assert cmd.execution_time == utime

    def test_request_battery_charge_sets_ra_dec_obsid(self, acs):
        ra, dec, obsid = 45.0, 30.0, 0xBEEF
        utime = 1514764800.0

        acs.request_battery_charge(utime, ra, dec, obsid)
        cmd = acs.command_queue[0]
        assert cmd.ra == ra and cmd.dec == dec and cmd.obsid == obsid

    def test_request_battery_charge_logs_info(self, acs):
        ra, dec, obsid = 45.0, 30.0, 0xBEEF
        utime = 1514764800.0

        acs.request_battery_charge(utime, ra, dec, obsid)
        # Test passes if no exception is raised - logging is tested via print statements

    def test_request_end_battery_charge_enqueues_command(self, acs):
        utime = 1514764800.0

        acs.request_end_battery_charge(utime)
        assert len(acs.command_queue) == 1

    def test_request_end_battery_charge_creates_correct_command_type(self, acs):
        utime = 1514764800.0

        acs.request_end_battery_charge(utime)
        cmd = acs.command_queue[0]
        assert cmd.command_type == ACSCommandType.END_BATTERY_CHARGE

    def test_request_end_battery_charge_sets_execution_time(self, acs):
        utime = 1514764800.0

        acs.request_end_battery_charge(utime)
        cmd = acs.command_queue[0]
        assert cmd.execution_time == utime

    def test_request_end_battery_charge_logs_info(self, acs):
        utime = 1514764800.0

        acs.request_end_battery_charge(utime)
        # Test passes if no exception is raised - logging is tested via print statements

    def test_initiate_emergency_charging_calls_emergency_module(self, acs):
        utime = 1514764800.0
        lastra, lastdec = 10.0, 20.0
        current_ppt = Mock()

        mock_emergency_charging = Mock()
        mock_charging_ppt = Mock()
        mock_charging_ppt.ra = 45.0
        mock_charging_ppt.dec = 30.0
        mock_charging_ppt.obsid = 0xC4A6
        mock_emergency_charging.initiate_emergency_charging = Mock(
            return_value=mock_charging_ppt
        )

        mock_ephem = Mock()

        with patch.object(acs, "request_battery_charge"):
            acs.initiate_emergency_charging(
                utime, mock_ephem, mock_emergency_charging, lastra, lastdec, current_ppt
            )
            mock_emergency_charging.initiate_emergency_charging.assert_called_once_with(
                utime, mock_ephem, lastra, lastdec, current_ppt
            )

    def test_initiate_emergency_charging_requests_charge(self, acs):
        utime = 1514764800.0
        lastra, lastdec = 10.0, 20.0
        current_ppt = Mock()

        mock_emergency_charging = Mock()
        mock_charging_ppt = Mock()
        mock_charging_ppt.ra = 45.0
        mock_charging_ppt.dec = 30.0
        mock_charging_ppt.obsid = 0xC4A6
        mock_emergency_charging.initiate_emergency_charging = Mock(
            return_value=mock_charging_ppt
        )

        mock_ephem = Mock()

        with patch.object(acs, "request_battery_charge") as mock_request:
            acs.initiate_emergency_charging(
                utime, mock_ephem, mock_emergency_charging, lastra, lastdec, current_ppt
            )
            mock_request.assert_called_once_with(utime, 45.0, 30.0, 0xC4A6)

    def test_initiate_emergency_charging_returns_updated_ra_dec_ppt(self, acs):
        utime = 1514764800.0
        lastra, lastdec = 10.0, 20.0
        current_ppt = Mock()

        mock_emergency_charging = Mock()
        mock_charging_ppt = Mock()
        mock_charging_ppt.ra = 45.0
        mock_charging_ppt.dec = 30.0
        mock_charging_ppt.obsid = 0xC4A6
        mock_emergency_charging.initiate_emergency_charging = Mock(
            return_value=mock_charging_ppt
        )

        mock_ephem = Mock()

        with patch.object(acs, "request_battery_charge"):
            ra, dec, ppt = acs.initiate_emergency_charging(
                utime, mock_ephem, mock_emergency_charging, lastra, lastdec, current_ppt
            )
            assert ra == 45.0

    def test_initiate_emergency_charging_returns_correct_dec(self, acs):
        utime = 1514764800.0
        lastra, lastdec = 10.0, 20.0
        current_ppt = Mock()

        mock_emergency_charging = Mock()
        mock_charging_ppt = Mock()
        mock_charging_ppt.ra = 45.0
        mock_charging_ppt.dec = 30.0
        mock_charging_ppt.obsid = 0xC4A6
        mock_emergency_charging.initiate_emergency_charging = Mock(
            return_value=mock_charging_ppt
        )

        mock_ephem = Mock()

        with patch.object(acs, "request_battery_charge"):
            ra, dec, ppt = acs.initiate_emergency_charging(
                utime, mock_ephem, mock_emergency_charging, lastra, lastdec, current_ppt
            )
            assert dec == 30.0

    def test_initiate_emergency_charging_returns_ppt(self, acs):
        utime = 1514764800.0
        lastra, lastdec = 10.0, 20.0
        current_ppt = Mock()

        mock_emergency_charging = Mock()
        mock_charging_ppt = Mock()
        mock_charging_ppt.ra = 45.0
        mock_charging_ppt.dec = 30.0
        mock_charging_ppt.obsid = 0xC4A6
        mock_emergency_charging.initiate_emergency_charging = Mock(
            return_value=mock_charging_ppt
        )

        mock_ephem = Mock()

        with patch.object(acs, "request_battery_charge"):
            ra, dec, ppt = acs.initiate_emergency_charging(
                utime, mock_ephem, mock_emergency_charging, lastra, lastdec, current_ppt
            )
            assert ppt == mock_charging_ppt

    def test_initiate_emergency_charging_failure_does_not_request(self, acs):
        utime = 1514764800.0
        lastra, lastdec = 10.0, 20.0
        current_ppt = Mock()

        mock_emergency_charging = Mock()
        mock_emergency_charging.initiate_emergency_charging = Mock(return_value=None)

        mock_ephem = Mock()

        with patch.object(acs, "request_battery_charge") as mock_request:
            acs.initiate_emergency_charging(
                utime, mock_ephem, mock_emergency_charging, lastra, lastdec, current_ppt
            )
            mock_request.assert_not_called()

    def test_initiate_emergency_charging_failure_returns_lastra(self, acs):
        utime = 1514764800.0
        lastra, lastdec = 10.0, 20.0
        current_ppt = Mock()

        mock_emergency_charging = Mock()
        mock_emergency_charging.initiate_emergency_charging = Mock(return_value=None)

        mock_ephem = Mock()

        with patch.object(acs, "request_battery_charge"):
            ra, dec, ppt = acs.initiate_emergency_charging(
                utime, mock_ephem, mock_emergency_charging, lastra, lastdec, current_ppt
            )
            assert ra == lastra

    def test_initiate_emergency_charging_failure_returns_lastdec(self, acs):
        utime = 1514764800.0
        lastra, lastdec = 10.0, 20.0
        current_ppt = Mock()

        mock_emergency_charging = Mock()
        mock_emergency_charging.initiate_emergency_charging = Mock(return_value=None)

        mock_ephem = Mock()

        with patch.object(acs, "request_battery_charge"):
            ra, dec, ppt = acs.initiate_emergency_charging(
                utime, mock_ephem, mock_emergency_charging, lastra, lastdec, current_ppt
            )
            assert dec == lastdec

    def test_initiate_emergency_charging_failure_returns_none_ppt(self, acs):
        utime = 1514764800.0
        lastra, lastdec = 10.0, 20.0
        current_ppt = Mock()

        mock_emergency_charging = Mock()
        mock_emergency_charging.initiate_emergency_charging = Mock(return_value=None)

        mock_ephem = Mock()

        with patch.object(acs, "request_battery_charge"):
            ra, dec, ppt = acs.initiate_emergency_charging(
                utime, mock_ephem, mock_emergency_charging, lastra, lastdec, current_ppt
            )
            assert ppt is None

    def test_start_battery_charge_executes_enqueue_command(self, acs):
        command = ACSCommand(
            command_type=ACSCommandType.START_BATTERY_CHARGE,
            execution_time=1514764800.0,
            ra=45.0,
            dec=30.0,
            obsid=0xBEEF,
        )

        with (
            patch.object(acs, "enqueue_command") as mock_enqueue_command,
            patch(
                "conops.Pointing.visibility",
                new=lambda self, *args, **kwargs: (
                    setattr(self, "windows", [[1514764800.0, 1514764900.0]]),
                    0,
                )[-1],
            ),
            patch("conops.Pointing.next_vis", return_value=1514764800.0),
        ):
            acs.config.spacecraft_bus.attitude_control.predict_slew.return_value = (
                0.0,
                (Mock(), Mock()),
            )
            acs.config.spacecraft_bus.attitude_control.slew_time.return_value = 10.0
            acs._start_battery_charge(command, 1514764800.0)
            # Check that enqueue_command was called (which will enqueue a SLEW_TO_TARGET command)
            assert mock_enqueue_command.call_count == 1
            enqueued_command = mock_enqueue_command.call_args[0][0]
            assert enqueued_command.command_type == ACSCommandType.SLEW_TO_TARGET
            assert enqueued_command.slew.endra == 45.0
            assert enqueued_command.slew.enddec == 30.0
            assert enqueued_command.slew.obsid == 0xBEEF
            assert enqueued_command.slew.obstype == "CHARGE"

    def test_start_battery_charge_logs(self, acs):
        command = ACSCommand(
            command_type=ACSCommandType.START_BATTERY_CHARGE,
            execution_time=1514764800.0,
            ra=45.0,
            dec=30.0,
            obsid=0xBEEF,
        )

        with (
            patch.object(acs, "enqueue_command"),
            patch(
                "conops.Pointing.visibility",
                new=lambda self, *args, **kwargs: (
                    setattr(self, "windows", [[1514764800.0, 1514764900.0]]),
                    0,
                )[-1],
            ),
            patch("conops.Pointing.next_vis", return_value=1514764800.0),
        ):
            acs.config.spacecraft_bus.attitude_control.predict_slew.return_value = (
                0.0,
                (Mock(), Mock()),
            )
            acs.config.spacecraft_bus.attitude_control.slew_time.return_value = 10.0
            acs._start_battery_charge(command, 1514764800.0)
            # Test passes if no exception is raised - logging is tested via print statements

    def test_start_battery_charge_missing_params_does_not_enqueue_command(self, acs):
        command = ACSCommand(
            command_type=ACSCommandType.START_BATTERY_CHARGE,
            execution_time=1514764800.0,
            ra=None,
            dec=None,
            obsid=None,
        )

        with patch.object(acs, "enqueue_command") as mock_enqueue_command:
            acs._start_battery_charge(command, 1514764800.0)
            mock_enqueue_command.assert_not_called()

    def test_end_battery_charge_calls_enqueue_command_with_last_ppt(self, acs):
        mock_ppt = Mock(spec=Slew)
        mock_ppt.endra = 45.0
        mock_ppt.enddec = 30.0
        mock_ppt.obsid = 100
        acs.last_ppt = mock_ppt

        with (
            patch.object(acs, "enqueue_command") as mock_enqueue_command,
            patch(
                "conops.Pointing.visibility",
                new=lambda self, *args, **kwargs: (
                    setattr(self, "windows", [[1514764800.0, 1514764900.0]]),
                    0,
                )[-1],
            ),
            patch("conops.Pointing.next_vis", return_value=1514764800.0),
        ):
            acs.config.spacecraft_bus.attitude_control.predict_slew.return_value = (
                0.0,
                (Mock(), Mock()),
            )
            acs.config.spacecraft_bus.attitude_control.slew_time.return_value = 10.0
            acs._end_battery_charge(1514764800.0)
            # Check that enqueue_command was called with a SLEW_TO_TARGET command
            assert mock_enqueue_command.call_count == 1
            enqueued_command = mock_enqueue_command.call_args[0][0]
            assert enqueued_command.command_type == ACSCommandType.SLEW_TO_TARGET
            assert enqueued_command.slew.endra == 45.0
            assert enqueued_command.slew.enddec == 30.0
            assert enqueued_command.slew.obsid == 100

    def test_end_battery_charge_no_last_ppt_does_not_enqueue_command(self, acs):
        acs.last_ppt = None

        with patch.object(acs, "enqueue_command") as mock_enqueue_command:
            acs._end_battery_charge(1514764800.0)
            mock_enqueue_command.assert_not_called()

    def test_end_battery_charge_no_last_ppt_logs(self, acs):
        acs.last_ppt = None

        # Test passes if no exception is raised - logging is tested via print statements
        acs._end_battery_charge(1514764800.0)

    def test_process_commands_calls_start_battery_charge(self, acs):
        command = ACSCommand(
            command_type=ACSCommandType.START_BATTERY_CHARGE,
            execution_time=1514764800.0,
            ra=45.0,
            dec=30.0,
            obsid=0xBEEF,
        )
        acs.command_queue = [command]

        with patch.object(acs, "_start_battery_charge") as mock_start:
            acs._process_commands(1514764800.0)
            mock_start.assert_called_once_with(command, 1514764800.0)

    def test_process_commands_calls_end_battery_charge(self, acs):
        command = ACSCommand(
            command_type=ACSCommandType.END_BATTERY_CHARGE,
            execution_time=1514764800.0,
        )
        acs.command_queue = [command]

        with patch.object(acs, "_end_battery_charge") as mock_end:
            acs._process_commands(1514764800.0)
            mock_end.assert_called_once_with(1514764800.0)



"""Unit tests for Pass and PassTimes classes."""

from datetime import datetime, timezone
from unittest.mock import Mock, patch

import pytest

from conops import (
    Constraint,
    GroundStationRegistry,
    Pass,
    PassTimes,
)


class TestPassInitialization:
    """Test Pass initialization."""

    def test_pass_creation_minimal(
        self, mock_config, mock_constraint, mock_acs_config, base_begin
    ):
        """Test creating a Pass with minimal parameters."""
        # Configure ACS and pass config via mock_config
        mock_config.spacecraft_bus.attitude_control = mock_acs_config
        p = Pass(
            config=mock_config,
            ephem=None,
            station="SGS",
            begin=base_begin,
            length=480.0,
        )
        assert p.station == "SGS"
        assert p.begin == base_begin
        assert p.length == 480.0
        assert p.obsid == 0xFFFF

    def test_pass_creation_full(
        self,
        mock_config,
        mock_constraint,
        mock_ephem,
        mock_acs_config,
        base_begin,
        default_length,
    ):
        """Test creating a Pass with all parameters."""
        mock_config.spacecraft_bus.attitude_control = mock_acs_config
        p = Pass(
            config=mock_config,
            ephem=mock_ephem,
            station="SGS",
            begin=base_begin,
            length=default_length,
            gsstartra=10.0,
            gsstartdec=20.0,
            gsendra=15.0,
            gsenddec=25.0,
        )
        assert p.station == "SGS"
        assert p.begin == base_begin
        assert p.length == default_length
        assert p.gsstartra == 10.0
        assert p.gsstartdec == 20.0
        assert p.gsendra == 15.0
        assert p.gsenddec == 25.0


class TestPassProperties:
    """Test Pass properties."""

    def test_ground_station_start_pointing(self, basic_pass):
        """Test ground station start pointing fields are set."""
        assert basic_pass.gsstartra == 10.0
        assert basic_pass.gsstartdec == 20.0

    def test_ground_station_end_pointing(self, basic_pass):
        """Test ground station end pointing fields are set."""
        assert basic_pass.gsendra == 15.0
        assert basic_pass.gsenddec == 25.0

    def test_end_property(self, basic_pass):
        """Test end property."""
        assert basic_pass.end == basic_pass.begin + basic_pass.length

    def test_end_property_no_length(
        self, mock_config, mock_constraint, mock_acs_config, base_begin
    ):
        """Test end property works when length is set."""
        mock_config.spacecraft_bus.attitude_control = mock_acs_config
        p = Pass(
            config=mock_config,
            ephem=None,
            station="SGS",
            begin=base_begin,
            length=480.0,
        )
        assert p.end == base_begin + 480.0


class TestPassMethods:
    """Test Pass methods."""

    def test_str_method(self, basic_pass):
        """Test __str__ method."""
        result = str(basic_pass)
        assert "SGS" in result
        assert "8.0 mins" in result

    def test_in_pass_true(self, basic_pass):
        """Test in_pass returns True during pass."""
        assert basic_pass.in_pass(1514764900.0) is True

    def test_in_pass_false_before(self, basic_pass):
        """Test in_pass returns False before pass."""
        assert basic_pass.in_pass(1514764700.0) is False

    def test_in_pass_false_after(self, basic_pass):
        """Test in_pass returns False after pass."""
        assert basic_pass.in_pass(1514765400.0) is False

    def test_time_to_pass_minutes(self, basic_pass):
        """Test time_to_pass returns minutes format."""
        with patch("time.time", return_value=1514764800.0 - 1800):  # 30 mins before
            result = basic_pass.time_to_pass()
            assert "30 mins" in result

    def test_time_to_pass_hours(self, basic_pass):
        """Test time_to_pass returns hours format."""
        with patch("time.time", return_value=1514764800.0 - 7200):  # 2 hours before
            result = basic_pass.time_to_pass()
            assert "hours" in result

    def test_ra_dec_before_pass(self, basic_pass, two_step_utime, small_ra, small_dec):
        """Test ra_dec with empty profile."""
        # ra_dec will fail with IndexError on empty lists, which is expected behavior
        # Tests using ra_dec should populate the arrays first
        basic_pass.utime = two_step_utime
        basic_pass.ra = small_ra
        basic_pass.dec = small_dec
        ra, dec = basic_pass.ra_dec(two_step_utime[0] - 600.0)  # Before first time
        assert ra == 10.0
        assert dec == 20.0

    def test_ra_dec_lookup(self, basic_pass, three_step_utime, small_ra, small_dec):
        """Test ra_dec returns values from profile."""
        basic_pass.utime = three_step_utime
        basic_pass.ra = [10.0, 12.0, 14.0]
        basic_pass.dec = [20.0, 22.0, 24.0]
        # ra_dec uses searchsorted and returns [idx] directly, so
        # at time 1514764900.0 (exact match), it returns index 1
        ra, dec = basic_pass.ra_dec(1514764900.0)
        assert ra == 12.0
        assert dec == 22.0


class TestPassTimeToSlew:
    """Test Pass.time_to_slew method."""

    def test_time_to_slew_no_pass_profile(self, basic_pass_mock):
        """Test time_to_slew returns False when pass has no ra/dec profile."""
        # Before pass with empty profile - should target start but profile is empty
        result = basic_pass_mock.time_to_slew(1514764700.0, ra=10.0, dec=20.0)
        assert result is False

    def test_time_to_slew_early_with_valid_profile(
        self, basic_pass_mock, start_ra, start_dec, two_step_utime
    ):
        """Test time_to_slew with valid profile and early timing."""
        basic_pass_mock.ra = start_ra  # Pass start pointing
        basic_pass_mock.dec = start_dec
        basic_pass_mock.utime = two_step_utime
        # Before pass at 1514764700.0, targeting pass start at [15, 25]
        result = basic_pass_mock.time_to_slew(1514764700.0, ra=10.0, dec=20.0)
        assert isinstance(result, bool)

    def test_time_to_slew_close_to_slew_time(
        self, basic_pass_mock, start_ra, start_dec, two_step_utime
    ):
        """Test time_to_slew when within ephemeris step buffer."""
        basic_pass_mock.ra = start_ra
        basic_pass_mock.dec = start_dec
        basic_pass_mock.utime = two_step_utime
        # Get very close to slew time - within step_size * 2 buffer
        result = basic_pass_mock.time_to_slew(1514764750.0, ra=10.0, dec=20.0)
        assert isinstance(result, bool)

    def test_time_to_slew_during_pass(
        self, basic_pass_mock, start_ra, start_dec, two_step_utime
    ):
        """Test time_to_slew when pass has already started - targets current position."""
        basic_pass_mock.ra = start_ra
        basic_pass_mock.dec = start_dec
        basic_pass_mock.utime = two_step_utime
        # During pass (utime >= begin), should target current pass position
        # This should work without raising
        result = basic_pass_mock.time_to_slew(1514764850.0, ra=10.0, dec=20.0)
        assert isinstance(result, bool)

    def test_time_to_slew_within_ephem_step_buffer_returns_true(
        self,
        mock_config,
        mock_constraint,
        create_ephem,
        acs_mock,
        base_begin,
    ):
        """time_to_slew should return True when time_until_slew is within ephem.step_size*2 buffer."""
        # Use a small real TLE ephemeris instance to satisfy pydantic validation
        begin_dt = datetime(2025, 8, 15, 0, 0, 0, tzinfo=timezone.utc)
        end_dt = datetime(2025, 8, 15, 0, 15, 0, tzinfo=timezone.utc)
        ephem = create_ephem(begin_dt, end_dt, step_size=60)

        mock_config.constraint = mock_constraint
        mock_config.spacecraft_bus.attitude_control = acs_mock
        p = Pass(
            config=mock_config,
            ephem=ephem,
            station="SGS",
            begin=base_begin,
            length=600.0,
        )
        # Provide a start profile so the "on time" branch has data to use
        p.utime = [1514764800.0, 1514764860.0]
        p.ra = [10.0, 12.0]
        p.dec = [20.0, 22.0]

        # utime = 1514764600 (200s before begin). With slew_time=100 -> time_until_slew = 100 <= 120 (buffer)
        result = p.time_to_slew(1514764600.0, ra=0.0, dec=0.0)
        assert result is True

    def test_time_to_slew_outside_ephem_step_buffer_returns_false(
        self,
        mock_config,
        mock_constraint,
        create_ephem,
        acs_mock,
        base_begin,
    ):
        """time_to_slew should return False when time_until_slew is outside the ephem.step_size*2 buffer."""
        acs = acs_mock
        acs.slew_time.return_value = (
            50.0  # smaller slewtime -> time_until_slew becomes larger than buffer
        )

        # Use a small real TLE ephemeris instance to satisfy pydantic validation
        begin_dt = datetime(2025, 8, 15, 0, 0, 0, tzinfo=timezone.utc)
        end_dt = datetime(2025, 8, 15, 0, 15, 0, tzinfo=timezone.utc)
        ephem = create_ephem(begin_dt, end_dt, step_size=60)

        mock_config.constraint = mock_constraint
        mock_config.spacecraft_bus.attitude_control = acs
        p = Pass(
            config=mock_config,
            ephem=ephem,
            station="SGS",
            begin=base_begin,
            length=600.0,
        )
        p.utime = [1514764800.0, 1514764860.0]
        p.ra = [10.0, 12.0]
        p.dec = [20.0, 22.0]

        # utime same as previous; with slew_time=50 -> time_until_slew = 150 > 120 (buffer)
        result = p.time_to_slew(1514764600.0, ra=0.0, dec=0.0)
        assert result is False

    def test_time_to_slew_raises_value_error_when_no_acs_config(
        self, mock_config, mock_constraint, create_ephem, base_begin
    ):
        """time_to_slew should raise ValueError when no ACS config is provided."""
        # Use a small real ephemeris to satisfy pydantic validation
        begin_dt = datetime(2025, 8, 15, 0, 0, 0, tzinfo=timezone.utc)
        end_dt = datetime(2025, 8, 15, 0, 15, 0, tzinfo=timezone.utc)
        ephem_obj = create_ephem(begin_dt, end_dt, step_size=60)

        mock_config.constraint = mock_constraint
        mock_config.spacecraft_bus.attitude_control = None
        p = Pass(
            config=mock_config,
            ephem=ephem_obj,
            station="SGS",
            begin=base_begin,
            length=600.0,
        )
        # Provide profile so we don't exit earlier
        p.utime = [1514764800.0, 1514764860.0]
        p.ra = [10.0, 12.0]
        p.dec = [20.0, 22.0]

        with pytest.raises(ValueError, match="ACS config must be set"):
            p.time_to_slew(1514764700.0, ra=0.0, dec=0.0)

    def test_time_to_slew_raises_assertion_if_no_ephemeris(
        self, mock_config, mock_constraint, mock_acs_config, base_begin
    ):
        """time_to_slew should assert if ephemeris is missing."""
        mock_config.constraint = mock_constraint
        mock_config.spacecraft_bus.attitude_control = mock_acs_config
        p = Pass(
            config=mock_config,
            ephem=None,
            station="SGS",
            begin=base_begin,
            length=600.0,
        )
        # Provide a simple profile so code proceeds past profile checks
        p.utime = [1514764800.0, 1514764860.0]
        p.ra = [10.0, 12.0]
        p.dec = [20.0, 22.0]

        with pytest.raises(
            AssertionError, match="Ephemeris must be set for Pass class"
        ):
            p.time_to_slew(1514764700.0, ra=0.0, dec=0.0)


class TestPassTimes:
    """Test PassTimes class."""

    def test_passtimes_initialization(self, mock_constraint, mock_config):
        """Test PassTimes initialization."""
        pt = PassTimes(config=mock_config)
        assert pt.constraint is mock_constraint
        assert pt.ephem is mock_constraint.ephem
        assert pt.config is mock_config
        assert pt.passes == []
        assert pt.length == 1
        assert pt.minelev == 10.0
        assert pt.minlen == 480
        assert pt.schedule_chance == 1.0

    def test_passtimes_uses_default_ground_stations(self, mock_constraint, mock_config):
        """Test PassTimes uses default ground stations when none provided."""
        mock_config.ground_stations = None
        pt = PassTimes(config=mock_config)
        assert isinstance(pt.ground_stations, GroundStationRegistry)

    def test_passtimes_uses_provided_ground_stations(
        self, mock_constraint, mock_config
    ):
        """Test PassTimes uses provided ground stations."""
        custom_gs = GroundStationRegistry()
        mock_config.ground_stations = custom_gs
        pt = PassTimes(config=mock_config)
        assert pt.ground_stations is custom_gs

    def test_passtimes_requires_ephemeris(self, mock_config):
        """Test PassTimes requires ephemeris."""
        constraint = Mock()
        constraint.ephem = None
        with pytest.raises(AssertionError, match="Ephemeris must be set"):
            # PassTimes reads constraint from config.constraint
            mock_config.constraint = constraint
            PassTimes(config=mock_config)

    def test_passtimes_getitem(self, mock_constraint, mock_config):
        """Test PassTimes __getitem__."""
        pt = PassTimes(config=mock_config)
        p1 = Mock()
        p2 = Mock()
        pt.passes = [p1, p2]
        assert pt[0] is p1
        assert pt[1] is p2

    def test_passtimes_len(self, mock_constraint, mock_config):
        """Test PassTimes __len__."""
        pt = PassTimes(config=mock_config)
        pt.passes = [Mock(), Mock(), Mock()]
        assert len(pt) == 3

    def test_next_pass_found(self, mock_constraint, mock_config):
        """Test next_pass returns next pass after given time."""
        pt = PassTimes(config=mock_config)
        p1 = Mock()
        p1.begin = 1000.0
        p2 = Mock()
        p2.begin = 2000.0
        p3 = Mock()
        p3.begin = 3000.0
        pt.passes = [p1, p2, p3]
        assert pt.next_pass(1500.0) is p2

    def test_next_pass_none(self, mock_constraint, mock_config):
        """Test next_pass returns None when no future passes."""
        pt = PassTimes(config=mock_config)
        p1 = Mock()
        p1.begin = 1000.0
        pt.passes = [p1]
        assert pt.next_pass(2000.0) is None

    def test_request_passes(self, mock_constraint, mock_config):
        """Test request_passes returns passes at requested rate."""
        pt = PassTimes(config=mock_config)
        # Create passes every ~4 hours
        for i in range(6):
            p = Mock()
            p.begin = i * 14400.0  # Every 4 hours
            pt.passes.append(p)

        with patch("numpy.random.random", return_value=0.5):
            scheduled = pt.request_passes(req_gsnum=6, gsprob=0.9)
            assert len(scheduled) > 0

    def test_request_passes_probability(self, mock_constraint, mock_config):
        """Test request_passes respects probability."""
        pt = PassTimes(config=mock_config)
        for i in range(10):
            p = Mock()
            p.begin = i * 20000.0
            pt.passes.append(p)

        # All random values > 0.9, should not schedule any
        with patch("numpy.random.random", return_value=0.95):
            scheduled = pt.request_passes(req_gsnum=10, gsprob=0.9)
            assert len(scheduled) == 0

    def test_get_requires_ephemeris(self, mock_constraint, mock_config):
        """Test PassTimes initialization requires ephemeris."""
        mock_constraint.ephem = None
        with pytest.raises(AssertionError, match="Ephemeris must be set"):
            PassTimes(config=mock_config)

    def test_get_sorts_passes_by_time(
        self, mock_constraint, mock_config, mock_ephemeris_100
    ):
        """Test get method sorts passes by time."""
        ephem = mock_ephemeris_100
        mock_constraint.ephem = ephem

        pt = PassTimes(config=mock_config)

        # Manually add passes out of order
        mock_config.constraint = mock_constraint
        mock_config.spacecraft_bus.attitude_control = Mock()
        p1 = Pass(
            config=mock_config,
            ephem=None,
            station="A",
            begin=3000.0,
            length=100.0,
        )
        mock_config.constraint = mock_constraint
        mock_config.spacecraft_bus.attitude_control = Mock()
        p2 = Pass(
            config=mock_config,
            ephem=None,
            station="B",
            begin=1000.0,
            length=100.0,
        )
        mock_config.constraint = mock_constraint
        mock_config.spacecraft_bus.attitude_control = Mock()
        p3 = Pass(
            config=mock_config,
            ephem=None,
            station="C",
            begin=2000.0,
            length=100.0,
        )
        pt.passes = [p1, p2, p3]

        # Sort
        pt.passes.sort(key=lambda x: x.begin, reverse=False)

        assert pt.passes[0].begin == 1000.0
        assert pt.passes[1].begin == 2000.0
        assert pt.passes[2].begin == 3000.0


class TestPassEdgeCases:
    """Test edge cases and error conditions."""

    def test_pass_with_empty_pointing_profile(self, basic_pass):
        """Test Pass with empty ra/dec lists."""
        basic_pass.utime = []
        basic_pass.ra = []
        basic_pass.dec = []
        # Empty profile is valid state
        assert len(basic_pass.ra) == 0
        assert len(basic_pass.utime) == 0

    def test_pass_obsid_default_value(
        self, mock_config, mock_constraint, mock_acs_config, base_begin
    ):
        """Test Pass obsid has correct default value."""
        mock_config.constraint = mock_constraint
        mock_config.spacecraft_bus.attitude_control = mock_acs_config
        p = Pass(
            config=mock_config,
            station="SGS",
            begin=base_begin,
            length=480.0,
        )
        assert p.obsid == 0xFFFF

    def test_pass_scheduling_fields(
        self, mock_config, mock_constraint, mock_acs_config, base_begin
    ):
        """Test Pass scheduling fields have correct defaults."""
        mock_config.constraint = mock_constraint
        mock_config.spacecraft_bus.attitude_control = mock_acs_config
        p = Pass(
            config=mock_config,
            station="SGS",
            begin=base_begin,
            length=480.0,
        )
        assert p.slewrequired == 0.0
        assert p.slewlate == 0.0


class TestPassTimesGetIntegration:
    """Integration tests for PassTimes.get method with real ephemeris."""

    def test_get_with_real_ephemeris(self, mock_config, create_ephem):
        """Test PassTimes.get with a real TLE ephemeris."""

        # Create ephemeris for a short time period
        begin = datetime(2025, 8, 15, 0, 0, 0, tzinfo=timezone.utc)
        end = datetime(2025, 8, 15, 2, 0, 0, tzinfo=timezone.utc)
        ephem = create_ephem(begin, end, step_size=60)

        # Create constraint with ephemeris
        constraint = Mock(spec=Constraint)
        constraint.ephem = ephem

        # Create PassTimes
        mock_config.constraint = constraint
        passtimes = PassTimes(config=mock_config)

        # Set minlen to be very high so no passes are created
        # This exercises the code without creating actual passes
        passtimes.minlen = 100000.0  # Very high minimum length

        # Run get method - this should execute all the code but filter out all passes
        passtimes.get(year=2025, day=227, length=1)

        # Verify passes list exists
        assert hasattr(passtimes, "passes")
        assert isinstance(passtimes.passes, list)
        # With such high minlen, no passes should be created
        assert len(passtimes.passes) == 0

    def test_get_creates_pass_objects(self, mock_config, create_ephem):
        """Test PassTimes.get actually creates Pass objects with realistic parameters."""

        # Create ephemeris for a longer time period to increase pass chances
        begin = datetime(2025, 8, 15, 0, 0, 0, tzinfo=timezone.utc)
        end = datetime(2025, 8, 16, 0, 0, 0, tzinfo=timezone.utc)

        ephem = create_ephem(begin, end, step_size=60)

        # Create constraint with ephemeris
        constraint = Mock(spec=Constraint)
        constraint.ephem = ephem

        # Create PassTimes
        mock_config.constraint = constraint
        passtimes = PassTimes(config=mock_config)

        # Set reasonable parameters
        passtimes.minlen = 60.0  # 1 minute minimum
        passtimes.minelev = 5.0  # 5 degrees minimum elevation
        passtimes.schedule_chance = 1.0  # Always schedule

        # Run get method
        passtimes.get(year=2025, day=227, length=1)

        # Verify passes list exists
        assert hasattr(passtimes, "passes")
        assert isinstance(passtimes.passes, list)
        # With a 12-hour period and low thresholds, we should get at least one pass
        # (This may be 0 if geometry doesn't work out, which is okay for coverage)
        assert len(passtimes.passes) >= 0


class TestPassTimesCurrent:
    """Tests for PassTimes.current_pass method."""

    def test_current_pass_returns_active_pass(self, mock_constraint, mock_config):
        """Should return the first pass whose in_pass returns True."""
        pt = PassTimes(config=mock_config)

        p1 = Mock()
        p1.in_pass.return_value = False
        p2 = Mock()
        p2.in_pass.return_value = True
        p3 = Mock()
        p3.in_pass.return_value = False

        pt.passes = [p1, p2, p3]
        assert pt.current_pass(1234.0) is p2

    def test_current_pass_none_when_no_active(self, mock_constraint, mock_config):
        """Should return None when no passes are active."""
        pt = PassTimes(config=mock_config)

        p1 = Mock()
        p1.in_pass.return_value = False
        p2 = Mock()
        p2.in_pass.return_value = False

        pt.passes = [p1, p2]
        assert pt.current_pass(1234.0) is None

    def test_current_pass_empty_list_returns_none(self, mock_constraint, mock_config):
        """Should return None if no passes are present."""
        pt = PassTimes(config=mock_config)
        pt.passes = []
        assert pt.current_pass(1234.0) is None

    def test_current_pass_prefers_first_matching(self, mock_constraint, mock_config):
        """If multiple passes report active, the first in the list should be returned."""
        pt = PassTimes(config=mock_config)

        p1 = Mock()
        p1.in_pass.return_value = True
        p2 = Mock()
        p2.in_pass.return_value = True

        pt.passes = [p1, p2]
        assert pt.current_pass(1234.0) is p1

    def test_current_pass_with_real_pass_objects(
        self, mock_constraint, mock_config, mock_acs_config
    ):
        """Integration-style test using real Pass objects and time ranges."""
        pt = PassTimes(config=mock_config)

        mock_config.constraint = mock_constraint
        mock_config.spacecraft_bus.attitude_control = mock_acs_config
        p1 = Pass(
            config=mock_config,
            station="A",
            begin=1000.0,
            length=200.0,
        )
        mock_config.constraint = mock_constraint
        mock_config.spacecraft_bus.attitude_control = mock_acs_config
        p2 = Pass(
            config=mock_config,
            station="B",
            begin=1500.0,
            length=200.0,
        )

        pt.passes = [p1, p2]

        # 1100 is inside p1
        assert pt.current_pass(1100.0) is p1
        # 1600 is inside p2
        assert pt.current_pass(1600.0) is p2
        # 900 is before any pass
        assert pt.current_pass(900.0) is None

    def test_current_pass_raises_if_pass_in_pass_list_has_no_length(
        self, mock_constraint, mock_config, mock_acs_config
    ):
        """Test that Pass objects have length properly set."""
        pt = PassTimes(config=mock_config)

        mock_config.constraint = mock_constraint
        mock_config.spacecraft_bus.attitude_control = mock_acs_config
        good_pass = Pass(
            config=mock_config,
            station="GOOD",
            begin=2000.0,
            length=100.0,
        )

        pt.passes = [good_pass]
        # Should work fine with proper length
        assert pt.current_pass(2000.0) is good_pass  # In pass



"""Additional tests to increase coverage for DITLLogStore."""

import tempfile
from pathlib import Path

from conops.common import ACSMode
from conops.ditl import DITLEvent, DITLLogStore


def test_add_events_batch_and_filters():
    with tempfile.TemporaryDirectory() as tmp:
        db = Path(tmp) / "batch.sqlite"
        store = DITLLogStore(db)
        run_id = "batch-run"
        events = [
            DITLEvent.from_utime(990.0, "INFO", "pre", None, None),
            DITLEvent.from_utime(1000.0, "PASS", "start", 1, ACSMode.PASS),
            DITLEvent.from_utime(1020.0, "OBSERVATION", "obs", 2, ACSMode.SCIENCE),
        ]
        store.add_events(run_id, events)

        # end_time filter
        early = store.fetch_events(run_id, end_time=1005.0)
        assert len(early) == 2
        assert [e.event_type for e in early] == ["INFO", "PASS"]

        # Specific type filter
        only_pass = store.fetch_events(run_id, event_type="PASS")
        assert len(only_pass) == 1
        assert only_pass[0].description == "start"

        # Empty runs listing after different run added
        store.add_event("other-run", DITLEvent.from_utime(1000.0, "INFO", "x"))
        runs = set(store.fetch_runs())
        assert "batch-run" in runs and "other-run" in runs

        # Close is safe to call multiple times
        store.close()
        store.close()



"""Test fixtures for thermal subsystem tests."""

import pytest

from conops import Heater, PowerDraw

"""Unit tests for conops.thermal.Heater"""


@pytest.fixture
def simple_power_draw():
    """Simple PowerDraw with only nominal power."""
    return PowerDraw(nominal_power=10.0, peak_power=15.0)


@pytest.fixture
def mode_based_power_draw():
    """PowerDraw with mode-specific power values."""
    return PowerDraw(
        nominal_power=10.0,
        peak_power=15.0,
        power_mode={0: 5.0, 1: 10.0, 2: 12.0, 3: 20.0},
    )


@pytest.fixture
def simple_heater(simple_power_draw):
    """Heater with simple power draw."""
    return Heater(name="Test Heater", power_draw=simple_power_draw)


@pytest.fixture
def mode_heater(mode_based_power_draw):
    """Heater with mode-based power draw."""
    return Heater(name="Mode Heater", power_draw=mode_based_power_draw)



"""Test fixtures for constraint subsystem tests."""

from datetime import datetime, timezone
from unittest.mock import Mock

import pytest

from conops import Constraint


@pytest.fixture
def constraint():
    """Fixture for a basic Constraint instance."""
    return Constraint()


@pytest.fixture
def constraint_with_ephem():
    """Fixture for a Constraint instance with mocked ephem."""
    c = Constraint()
    c.ephem = Mock()
    c.ephem._tle_ephem = Mock()
    return c


@pytest.fixture
def time_list():
    """Fixture for a list of datetime objects."""
    return [
        datetime.fromtimestamp(1700000000.0, tz=timezone.utc),
        datetime.fromtimestamp(1700000100.0, tz=timezone.utc),
    ]



from .common import (
    dtutcfromtimestamp,
    givename,
    ics_date_conv,
    unixtime2date,
    unixtime2yearday,
)
from .enums import ACSCommandType, ACSMode, AntennaType, ChargeState, Polarization
from .vector import (
    angular_separation,
    great_circle,
    radec2vec,
    roll_over_angle,
    rotvec,
    scbodyvector,
    separation,
)

__all__ = [
    "ACSCommandType",
    "ACSMode",
    "AntennaType",
    "Polarization",
    "ChargeState",
    "dtutcfromtimestamp",
    "givename",
    "great_circle",
    "ics_date_conv",
    "radec2vec",
    "roll_over_angle",
    "rotvec",
    "scbodyvector",
    "separation",
    "unixtime2date",
    "unixtime2yearday",
    "angular_separation",
]



"""Tests for conops.config module."""

import json
from unittest.mock import Mock

from conops import (
    Battery,
    Constraint,
    FaultManagement,
    GroundStationRegistry,
    MissionConfig,
    Payload,
    SolarPanelSet,
    SpacecraftBus,
)


class TestConfig:
    """Test Config class initialization."""

    def test_config_sets_name(self, minimal_config):
        """Test that Config sets the provided name."""
        assert minimal_config["config"].name == "Test Config"

    def test_config_sets_spacecraft_bus(self, minimal_config):
        """Test that Config sets spacecraft_bus correctly."""
        assert (
            minimal_config["config"].spacecraft_bus == minimal_config["spacecraft_bus"]
        )

    def test_config_sets_solar_panel(self, minimal_config):
        """Test that Config sets solar_panel correctly."""
        assert minimal_config["config"].solar_panel == minimal_config["solar_panel"]

    def test_config_sets_payload(self, minimal_config):
        """Test that Config sets payload correctly."""
        assert minimal_config["config"].payload == minimal_config["payload"]

    def test_config_sets_battery(self, minimal_config):
        """Test that Config sets battery correctly."""
        assert minimal_config["config"].battery == minimal_config["battery"]

    def test_config_sets_constraint(self, minimal_config):
        """Test that Config sets constraint correctly."""
        assert minimal_config["config"].constraint == minimal_config["constraint"]

    def test_config_sets_ground_stations(self, minimal_config):
        """Test that Config sets ground_stations correctly."""
        assert (
            minimal_config["config"].ground_stations
            == minimal_config["ground_stations"]
        )

    def test_config_default_name(self):
        """Test that Config uses default name."""
        spacecraft_bus = Mock(spec=SpacecraftBus)
        solar_panel = Mock(spec=SolarPanelSet)
        payload = Mock(spec=Payload)
        battery = Mock(spec=Battery)
        battery.max_depth_of_discharge = 0.3  # Configure mock battery
        constraint = Mock(spec=Constraint)
        panel_constraint_mock = Mock()
        panel_constraint_mock.solar_panel = None
        constraint.panel_constraint = panel_constraint_mock
        ground_stations = Mock(spec=GroundStationRegistry)

        config = MissionConfig(
            spacecraft_bus=spacecraft_bus,
            solar_panel=solar_panel,
            payload=payload,
            battery=battery,
            constraint=constraint,
            ground_stations=ground_stations,
        )

        assert config.name == "Default Config"

    def test_config_sets_fault_management(self, minimal_config):
        """Test that Config sets fault_management correctly."""
        assert minimal_config["config"].fault_management is not None
        assert isinstance(minimal_config["config"].fault_management, FaultManagement)

    def test_init_fault_management_defaults_none(self):
        """Test init_fault_management_defaults does nothing if fault_management is None."""
        battery = Mock(spec=Battery)
        battery.max_depth_of_discharge = 0.3  # Configure mock battery
        config = MissionConfig(
            spacecraft_bus=Mock(spec=SpacecraftBus),
            solar_panel=Mock(spec=SolarPanelSet),
            payload=Mock(spec=Payload),
            battery=battery,
            constraint=Mock(spec=Constraint),
            ground_stations=Mock(spec=GroundStationRegistry),
        )
        # Set fault_management to None after creation to test the validator
        config.fault_management = None
        config.init_fault_management_defaults()
        # No assertions needed, just ensure no errors

    def test_init_fault_management_defaults_adds_threshold(self):
        """Test init_fault_management_defaults adds battery_level threshold if not present."""
        fault_management = FaultManagement()
        battery = Mock(spec=Battery)
        battery.max_depth_of_discharge = 0.2
        config = MissionConfig(
            spacecraft_bus=Mock(spec=SpacecraftBus),
            solar_panel=Mock(spec=SolarPanelSet),
            payload=Mock(spec=Payload),
            battery=battery,
            constraint=Mock(spec=Constraint),
            ground_stations=Mock(spec=GroundStationRegistry),
            fault_management=fault_management,
        )
        config.init_fault_management_defaults()
        assert any(t.name == "battery_level" for t in fault_management.thresholds)
        threshold = next(
            t for t in fault_management.thresholds if t.name == "battery_level"
        )
        assert threshold.yellow == 0.8
        assert (
            abs(threshold.red - 0.7) < 1e-10
        )  # Use approximate comparison for floating point

    def test_init_fault_management_defaults_threshold_exists(self):
        """Test init_fault_management_defaults does not add threshold if already present."""
        fault_management = FaultManagement()
        fault_management.add_threshold(
            "battery_level", yellow=0.5, red=0.4, direction="below"
        )
        battery = Mock(spec=Battery)
        battery.max_depth_of_discharge = 0.2
        config = MissionConfig(
            spacecraft_bus=Mock(spec=SpacecraftBus),
            solar_panel=Mock(spec=SolarPanelSet),
            payload=Mock(spec=Payload),
            battery=battery,
            constraint=Mock(spec=Constraint),
            ground_stations=Mock(spec=GroundStationRegistry),
            fault_management=fault_management,
        )
        config.init_fault_management_defaults()
        # Should have battery_level (that we added) and recorder_fill_fraction (added by init)
        assert len(fault_management.thresholds) == 2
        assert any(t.name == "battery_level" for t in fault_management.thresholds)
        assert any(
            t.name == "recorder_fill_fraction" for t in fault_management.thresholds
        )
        # Battery level should have our custom values, not defaults
        battery_threshold = next(
            t for t in fault_management.thresholds if t.name == "battery_level"
        )
        assert battery_threshold.yellow == 0.5
        assert battery_threshold.red == 0.4

    def test_from_json_file(self, tmp_path):
        """Test loading Config from JSON file."""
        json_data = {
            "name": "Test Config",
            "spacecraft_bus": {"mass": 100.0},
            "solar_panel": {"area": 10.0},
            "payload": {"mass": 50.0},
            "battery": {"capacity": 200.0, "max_depth_of_discharge": 0.8},
            "constraint": {"some_constraint": "value"},
            "ground_stations": {},
        }
        file_path = tmp_path / "config.json"
        with open(file_path, "w") as f:
            json.dump(json_data, f)
        config = MissionConfig.from_json_file(str(file_path))
        assert config.name == "Test Config"
        assert config.fault_management is not None

    def test_to_json_file(self, tmp_path):
        """Test saving Config to JSON file."""
        config = MissionConfig(
            name="Test Config",
            spacecraft_bus=Mock(spec=SpacecraftBus, mass=100.0),
            solar_panel=Mock(spec=SolarPanelSet, area=10.0),
            payload=Mock(spec=Payload, mass=50.0),
            battery=Mock(spec=Battery, capacity=200.0, max_depth_of_discharge=0.8),
            constraint=Mock(spec=Constraint),
            ground_stations=Mock(spec=GroundStationRegistry),
        )
        file_path = tmp_path / "config.json"
        config.to_json_file(str(file_path))
        assert file_path.exists()
        with open(file_path) as f:
            data = json.load(f)
        assert data["name"] == "Test Config"



"""Test fixtures for ditl subsystem tests."""

import datetime
from unittest.mock import Mock, patch

import numpy as np
import pytest
from matplotlib import pyplot as plt

from conops import DITL, ACSMode
from conops.ditl.ditl_mixin import DITLMixin


class DummyEphemeris:
    """Minimal mock ephemeris for testing."""

    def __init__(self, num_steps: int = 5):
        from datetime import datetime, timezone

        self.step_size = 60  # Use 60 second steps for faster tests
        # Create a short simulation: just a few minutes instead of 24 hours
        start_time = 1543276800  # 2018-11-27 00:00:00 UTC
        unix_times = np.arange(
            start_time, start_time + num_steps * self.step_size, self.step_size
        )
        self.timestamp = [
            datetime.fromtimestamp(float(t), tz=timezone.utc) for t in unix_times
        ]
        self.utime = unix_times
        # Add earth and sun attributes for ACS initialization
        self.earth = [Mock(ra=Mock(deg=0.0), dec=Mock(deg=0.0)) for _ in unix_times]
        self.sun = [Mock(ra=Mock(deg=45.0), dec=Mock(deg=23.5)) for _ in unix_times]

    def index(self, time):
        """Mock index method."""
        return 0


@pytest.fixture
def mock_config():
    """Create a minimal mock config with required attributes for DITLMixin."""
    cfg = Mock()
    cfg.name = "test"
    cfg.constraint = Mock()
    cfg.constraint.ephem = Mock()  # DITLMixin asserts this is not None
    cfg.constraint.ephem.earth = [Mock(ra=Mock(deg=0.0), dec=Mock(deg=0.0))]
    # Set timestamp to match DummyEphemeris range
    cfg.constraint.ephem.timestamp = [
        datetime.datetime(2018, 11, 27, 0, 0, 0, tzinfo=datetime.timezone.utc),
        datetime.datetime(2018, 11, 28, 0, 0, 0, tzinfo=datetime.timezone.utc),
    ]
    cfg.battery = Mock()
    cfg.battery.max_depth_of_discharge = 0.5
    return cfg


@pytest.fixture
def mock_ephem():
    """Create a mock ephemeris object."""
    return DummyEphemeris()


@pytest.fixture
def mock_config_detailed():
    """Create a mock config with all required subsystems."""
    config = Mock()

    # Mock constraint
    config.constraint = Mock()
    config.constraint.ephem = DummyEphemeris()
    config.constraint.panel_constraint = Mock()
    config.constraint.panel_constraint.solar_panel = Mock()
    config.constraint.in_constraint = Mock(return_value=False)

    # Mock battery
    config.battery = Mock()
    config.battery.battery_level = 0.8
    config.battery.battery_alert = False
    config.battery.drain = Mock()
    config.battery.charge = Mock()
    config.battery.panel_charge_rate = 100.0

    # Mock spacecraft bus
    config.spacecraft_bus = Mock()
    config.spacecraft_bus.power = Mock(return_value=50.0)
    config.spacecraft_bus.attitude_control = Mock()
    config.spacecraft_bus.attitude_control.predict_slew = Mock(return_value=(45.0, []))
    config.spacecraft_bus.attitude_control.slew_time = Mock(return_value=100.0)

    # Mock payload
    config.payload = Mock()
    config.payload.power = Mock(return_value=30.0)
    config.payload.data_generated = Mock(return_value=0.0)

    # Mock recorder
    config.recorder = Mock()
    config.recorder.current_volume_gb = 0.0
    config.recorder.get_fill_fraction = Mock(return_value=0.0)
    config.recorder.get_alert_level = Mock(return_value="none")
    config.recorder.add_data = Mock()
    config.recorder.remove_data = Mock(return_value=0.0)

    # Mock solar panel
    config.solar_panel = Mock()
    config.solar_panel.power = Mock(return_value=100.0)
    config.solar_panel.panel_illumination_fraction = Mock(return_value=0.5)
    config.solar_panel.illumination_and_power = Mock(return_value=(0.5, 100.0))
    config.solar_panel.optimal_charging_pointing = Mock(return_value=(45.0, 23.5))

    # Mock ground stations
    config.ground_stations = Mock()

    return config


@pytest.fixture
def ditl(mock_config_detailed, mock_ephem):
    """Create a DITL instance with mocked dependencies."""
    with (
        patch("conops.PassTimes") as mock_passtimes,
        patch("conops.ACS") as mock_acs_class,
    ):
        # Mock PassTimes
        mock_pt = Mock()
        mock_pt.passes = []
        mock_pt.get = Mock()
        mock_passtimes.return_value = mock_pt

        # Mock ACS
        mock_acs = Mock()
        mock_acs.ephem = None
        mock_acs.slewing = False
        mock_acs.inpass = False
        mock_acs.saa = None
        mock_acs.pointing = Mock(return_value=(0.0, 0.0, 0.0, 0))
        mock_acs.add_slew = Mock()
        mock_acs.passrequests = mock_pt
        mock_acs.get_mode = Mock(return_value=ACSMode.SCIENCE)
        mock_acs_class.return_value = mock_acs

        ditl = DITL(config=mock_config_detailed)
        ditl.ephem = mock_ephem
        ditl.acs = mock_acs
        ditl.plan = Mock()
        ditl.plan.which_ppt = Mock(
            return_value=Mock(ra=0.0, dec=0.0, obsid=1, obstype="science")
        )

        return ditl


@pytest.fixture
def mock_pass_inst():
    """Mock PassTimes instance."""
    return Mock()


@pytest.fixture
def mock_acs_inst():
    """Mock ACS instance."""
    return Mock()


@pytest.fixture
def ditl_instance(mock_config, mock_pass_inst, mock_acs_inst):
    """Fixture to create a DITLMixin instance with mocked dependencies."""
    with (
        patch("conops.ditl.ditl_mixin.PassTimes") as mock_pass_class,
        patch("conops.ditl.ditl_mixin.ACS") as mock_acs_class,
        patch("conops.ditl.ditl_mixin.Plan") as mock_plan_class,
    ):
        # Set return values for patched classes
        mock_pass_class.return_value = mock_pass_inst
        mock_acs_class.return_value = mock_acs_inst
        mock_plan_class.return_value = Mock()

        ditl = DITLMixin(config=mock_config)
        return ditl, mock_pass_inst, mock_acs_inst


@pytest.fixture
def populated_ditl(ditl_instance):
    """Fixture to populate DITLMixin with sample data for plotting."""
    ditl, _, _ = ditl_instance
    # Populate data arrays of same length
    base_time = 1514764800.0
    ditl.utime = [base_time + i * 60 for i in range(4)]
    ditl.ra = [1.0, 2.0, 3.0, 4.0]
    ditl.dec = [0.5, 0.6, 0.7, 0.8]
    ditl.mode = [0, 1, 2, 3]
    ditl.batterylevel = [0.2, 0.3, 0.4, 0.5]
    ditl.panel = [0.1, 0.2, 0.3, 0.4]
    ditl.power = [5.0, 6.0, 7.0, 8.0]
    ditl.obsid = [0, 1, 2, 3]
    return ditl


@pytest.fixture
def comprehensive_ditl(ditl_instance, mock_config):
    """Fixture to populate DITLMixin with comprehensive data for statistics."""
    from conops.ditl.ditl_stats import DITLStats

    # Create a class that inherits from both mixins
    class TestDITLWithStats(DITLMixin, DITLStats):
        pass

    ditl = TestDITLWithStats(config=mock_config)
    # Copy the initialized attributes from the ditl_instance
    original_ditl, _, _ = ditl_instance
    ditl.passes = original_ditl.passes
    ditl.acs = original_ditl.acs
    ditl.plan = original_ditl.plan
    ditl.executed_passes = original_ditl.executed_passes
    ditl.begin = original_ditl.begin
    ditl.end = original_ditl.end
    ditl.step_size = original_ditl.step_size

    # Mock battery capacity
    mock_config.battery.watthour = 100.0
    mock_config.battery.max_depth_of_discharge = 0.1

    # Mock recorder capacity
    mock_config.recorder.capacity_gb = 100.0
    mock_config.recorder.yellow_threshold = 0.8
    mock_config.recorder.red_threshold = 0.95

    # Populate comprehensive telemetry data
    base_time = 1514764800.0  # 2018-01-01 00:00:00 UTC
    ditl.utime = [base_time + i * 60 for i in range(10)]
    ditl.mode = [0, 1, 2, 0, 1, 2, 0, 1, 2, 0]  # Mix of modes
    ditl.obsid = [0, 10000, 10001, 0, 10002, 10003, 0, 10004, 10005, 0]
    ditl.ra = [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0]
    ditl.dec = [0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.1, 1.2, 1.3, 1.4]
    ditl.roll = [10.0, 11.0, 12.0, 13.0, 14.0, 15.0, 16.0, 17.0, 18.0, 19.0]
    ditl.batterylevel = [0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 0.8, 0.7]
    ditl.charge_state = [0, 1, 1, 0, 1, 1, 0, 1, 1, 0]
    ditl.power = [5.0, 6.0, 7.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0, 14.0]
    ditl.power_bus = [2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0, 5.5, 6.0, 6.5]
    ditl.power_payload = [3.0, 3.5, 4.0, 4.5, 5.0, 5.5, 6.0, 6.5, 7.0, 7.5]
    ditl.panel_power = [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0]
    ditl.panel = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]

    # Data management
    ditl.recorder_volume_gb = [0.0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5]
    ditl.recorder_fill_fraction = [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
    ditl.recorder_alert = [0, 0, 0, 1, 1, 1, 2, 2, 2, 2]
    ditl.data_generated_gb = [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
    ditl.data_downlinked_gb = [
        0.0,
        0.05,
        0.1,
        0.15,
        0.2,
        0.25,
        0.3,
        0.35,
        0.4,
        0.45,
    ]

    # Mock queue and ACS commands for statistics
    ditl.queue = Mock()
    ditl.queue.targets = [Mock(done=True), Mock(done=False), Mock(done=True)]

    ditl.acs = Mock()
    mock_cmd1 = Mock()
    mock_cmd1.command_type = Mock()
    mock_cmd1.command_type.name = "SLEW"
    mock_cmd2 = Mock()
    mock_cmd2.command_type = Mock()
    mock_cmd2.command_type.name = "OBSERVE"
    ditl.acs.commands = [mock_cmd1, mock_cmd2, mock_cmd1]

    # Mock executed passes
    mock_pass = Mock()
    mock_pass.begin = base_time + 120
    mock_pass.end = base_time + 240
    ditl.executed_passes = Mock()
    ditl.executed_passes.passes = [mock_pass]

    return ditl


@pytest.fixture
def statistics_output(comprehensive_ditl):
    """Fixture to capture output of print_statistics."""
    import io
    from contextlib import redirect_stdout

    f = io.StringIO()
    with redirect_stdout(f):
        comprehensive_ditl.print_statistics()
    return f.getvalue()


@pytest.fixture
def plot_figure(populated_ditl):
    """Fixture to create and yield the plot figure."""
    plt.close("all")
    with patch("matplotlib.pyplot.show"):
        populated_ditl.plot()
    fig = plt.gcf()
    yield fig
    plt.close("all")


@pytest.fixture
def ditl_with_payload_and_recorder(ditl_instance):
    """Fixture to set up ditl with payload and recorder mocks."""
    ditl, _, _ = ditl_instance
    ditl.payload = Mock()
    ditl.payload.data_generated.return_value = 0.1
    ditl.recorder = Mock()
    ditl.recorder.add_data.return_value = None
    ditl.recorder.remove_data.return_value = 0.05
    return ditl


@pytest.fixture
def ditl_with_pass_setup(ditl_instance, mock_config):
    """Fixture to set up ditl with pass-related mocks."""
    ditl, _, _ = ditl_instance
    ditl.payload = Mock()
    ditl.payload.data_generated.return_value = 0.1
    ditl.recorder = Mock()
    ditl.recorder.add_data.return_value = None
    ditl.recorder.remove_data.return_value = 0.05
    mock_pass = Mock()
    mock_pass.in_pass.return_value = True
    mock_pass.station = "station1"
    # No spacecraft comms config: set the pass config attribute like the real Pass
    mock_pass.config = Mock()
    mock_pass.config.spacecraft_bus = Mock()
    mock_pass.config.spacecraft_bus.communications = None
    ditl.acs = Mock()
    ditl.acs.passrequests = Mock()
    ditl.acs.passrequests.passes = [mock_pass]
    mock_station = Mock()
    # New per-band API: when no spacecraft comms, use GS overall max
    mock_station.get_overall_max_downlink.return_value = 100.0
    # Set up station to avoid the iterable error
    mock_station.bands = None  # This will make it use the overall max path
    ditl.config.ground_stations = Mock()
    ditl.config.ground_stations.get = Mock(return_value=mock_station)
    # Ensure spacecraft_bus.communications is None to use overall max
    if not hasattr(ditl.config, "spacecraft_bus"):
        ditl.config.spacecraft_bus = Mock()
    ditl.config.spacecraft_bus.communications = None
    return ditl
